from cmath import phase
from copy import deepcopy
from typing import Callable, Optional, Type

import pytest
from neuraxle.data_container import ARG_X_INPUTTED, ARG_Y_PREDICTD, IDT
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.callbacks import (CallbackList, EarlyStoppingCallback,
                                        MetricCallback)
from neuraxle.metaopt.data.aggregates import (BaseAggregate, Client,
                                              MetricResults, Project, Root,
                                              Round, Trial, TrialSplit,
                                              aggregate_2_dataclass,
                                              aggregate_2_subaggregate)
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT,
                                           AutoMLContext, BaseDataclass,
                                           BaseHyperparameterOptimizer,
                                           MetricResultsDataclass,
                                           ScopedLocation, dataclass_2_id_attr,
                                           dataclass_2_subdataclass)
from neuraxle.metaopt.validation import GridExplorationSampler
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import AddN, MultiplyByN
from sklearn.metrics import median_absolute_error
from testing.metaopt.test_automl_dataclasses import (SOME_FULL_SCOPED_LOCATION,
                                                     SOME_ROOT_DATACLASS)
from testing.metaopt.test_automl_repositories import CX_WITH_REPO_CTORS, TmpDir


class SomeException(Exception):
    pass


def _raise_if_is_at_level(aggregate: BaseAggregate, level_to_raise: Type[BaseAggregate], current_phase: str, phase_to_raise: str):
    if isinstance(aggregate, level_to_raise) and phase_to_raise == current_phase:
        raise SomeException(f"==> Raising at `{phase_to_raise}`of `{level_to_raise.__name__}` as planned. <==\n"
                            f"(This should not be catched).")
    else:
        return


BEGIN = "beginning"
END = "the end"


@pytest.mark.parametrize('lvlno, level_to_raise', list(enumerate(list(aggregate_2_subaggregate.keys())[1:])))
@pytest.mark.parametrize('phase_to_raise', [BEGIN, END])
@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_aggregate_exceptions_will_raise(
    lvlno: int, level_to_raise: Type[BaseAggregate],
    phase_to_raise: str,
    cx_repo_ctor: Callable[[Optional[TmpDir]], AutoMLContext],
):

    with pytest.raises(SomeException):
        root: Root = Root.vanilla(cx_repo_ctor())
        _raise_if_is_at_level(root, level_to_raise, BEGIN, phase_to_raise)
        with root.default_project() as ps:
            ps: Project = ps
            _raise_if_is_at_level(ps, level_to_raise, BEGIN, phase_to_raise)
            with ps.default_client() as cs:
                cs: Client = cs
                _raise_if_is_at_level(cs, level_to_raise, BEGIN, phase_to_raise)
                with cs.new_round(main_metric_name="some_metric") as rs:
                    rs: Round = rs.with_optimizer(GridExplorationSampler(1), HyperparameterSpace())
                    _raise_if_is_at_level(rs, level_to_raise, BEGIN, phase_to_raise)
                    with rs.new_rvs_trial() as ts:
                        ts: Trial = ts
                        _raise_if_is_at_level(ts, level_to_raise, BEGIN, phase_to_raise)
                        with ts.new_validation_split(continue_loop_on_error=False) as tss:
                            tss: TrialSplit = tss.with_n_epochs(1)
                            _raise_if_is_at_level(tss, level_to_raise, BEGIN, phase_to_raise)
                            with tss.managed_metric(rs.main_metric_name, True) as mrs:
                                mrs: MetricResults = mrs
                                _raise_if_is_at_level(mrs, level_to_raise, BEGIN, phase_to_raise)

                                mrs.add_train_result(0.0)
                                mrs.add_valid_result(0.0)

                                _raise_if_is_at_level(mrs, level_to_raise, END, phase_to_raise)
                            _raise_if_is_at_level(tss, level_to_raise, END, phase_to_raise)
                        _raise_if_is_at_level(ts, level_to_raise, END, phase_to_raise)
                    _raise_if_is_at_level(rs, level_to_raise, END, phase_to_raise)
                _raise_if_is_at_level(cs, level_to_raise, END, phase_to_raise)
            _raise_if_is_at_level(ps, level_to_raise, END, phase_to_raise)
        _raise_if_is_at_level(root, level_to_raise, END, phase_to_raise)
    pass


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_scoped_cascade_does_the_right_logging(tmpdir, cx_repo_ctor: Callable[[Optional[TmpDir]], AutoMLContext]):
    dact_train: DACT[IDT, ARG_X_INPUTTED, ARG_Y_PREDICTD] = DACT(
        ids=list(range(0, 10)), di=list(range(0, 10)), eo=list(range(100, 110)))
    dact_valid: DACT[IDT, ARG_X_INPUTTED, ARG_Y_PREDICTD] = DACT(
        ids=list(range(10, 20)), di=list(range(10, 20)), eo=list(range(110, 120)))
    hp_optimizer: BaseHyperparameterOptimizer = GridExplorationSampler(expected_n_trials=1)
    n_epochs = 3
    n_early_stopping_sentinel = 1
    n_effective_epochs = min(n_epochs, 1 + n_early_stopping_sentinel)
    callbacks = CallbackList([
        MetricCallback('MAE', median_absolute_error, False),
        EarlyStoppingCallback(n_early_stopping_sentinel, 'MAE')
    ])
    expected_scope = ScopedLocation(
        project_name=DEFAULT_PROJECT,
        client_name=DEFAULT_CLIENT,
        round_number=0,
        trial_number=0,
        split_number=0,
        metric_name='MAE',
    )
    p = Pipeline([
        MultiplyByN().with_hp_range(range(1, 3)),
        AddN().with_hp_range(range(99, 103)),
    ])
    hps: HyperparameterSpace = p.get_hyperparams_space()
    root: Root = Root.vanilla(cx_repo_ctor())

    with root.default_project() as ps:
        ps: Project = ps
        with ps.default_client() as cs:
            cs: Client = cs
            with cs.new_round(main_metric_name='MAE') as rs:
                rs: Round = rs.with_optimizer(hp_optimizer=hp_optimizer, hp_space=hps)
                with rs.new_rvs_trial() as ts:
                    ts: Trial = ts
                    with ts.new_validation_split(continue_loop_on_error=False) as tss:
                        tss: TrialSplit = tss.with_n_epochs(n_epochs)

                        for _ in range(n_epochs):
                            e = tss.next_epoch()

                            p = p.handle_fit(
                                dact_train.copy(),
                                tss.train_context())

                            eval_dact_train = p.handle_predict(
                                dact_train.without_eo(),
                                tss.validation_context())
                            eval_dact_valid = p.handle_predict(
                                dact_valid.without_eo(),
                                tss.validation_context())

                            if callbacks.call(
                                tss,
                                eval_dact_train.with_eo(dact_train.eo),
                                eval_dact_valid.with_eo(dact_valid.eo),
                                e == n_epochs
                            ):
                                break

    for dc_type in dataclass_2_id_attr.keys():
        dc: BaseDataclass = root.repo.load(expected_scope[:dc_type])
        assert isinstance(dc, dc_type)
        assert dc.get_id() == expected_scope[dc_type]

        if dc_type == MetricResultsDataclass:
            assert len(dc.get_sublocation()) == n_effective_epochs
        else:
            assert len(dc.get_sublocation()) == 1


@pytest.mark.parametrize("_type", list(aggregate_2_dataclass.keys())[:-1])
def test_ensure_consistent_subtypes_lozenge(_type: BaseAggregate):
    # subdataclass from dataclass_2_subdataclass:
    _dataclass = aggregate_2_dataclass[_type]
    _subdataclass_dataclass = dataclass_2_subdataclass[_dataclass]
    # subdataclass from aggregate_2_subaggregate (lozenge):
    _subaggregate = aggregate_2_subaggregate[_type]
    _subdataclass_agg = aggregate_2_dataclass[_subaggregate]

    # pre-requisite type side assertions:
    assert issubclass(_type, BaseAggregate)
    assert issubclass(_subaggregate, BaseAggregate)
    assert issubclass(_dataclass, BaseDataclass)
    assert issubclass(_subdataclass_agg, BaseDataclass)
    assert issubclass(_subdataclass_dataclass, BaseDataclass)
    # final lozenge consistency assertion:
    assert _subdataclass_dataclass == _subdataclass_agg


cant_be_shallow_aggs = [
    TrialSplit,
    MetricResults,
]


@pytest.mark.parametrize("aggregate_class", list(aggregate_2_subaggregate.keys())[1:])
@pytest.mark.parametrize("is_deep", [True, False])
@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_aggregates_creation(
    aggregate_class: Type[BaseAggregate],
    is_deep: bool,
    cx_repo_ctor: Callable[[Optional[TmpDir]], AutoMLContext]
):
    # Create repo from deep root DC:
    dataclass_class: Type = aggregate_2_dataclass[aggregate_class]
    scoped_loc: ScopedLocation = SOME_FULL_SCOPED_LOCATION[:dataclass_class]
    context = cx_repo_ctor()
    context.repo.save(SOME_ROOT_DATACLASS, ScopedLocation(), deep=True)

    # Retrieve scoped DC from root
    dataclass = SOME_ROOT_DATACLASS[scoped_loc]
    if not is_deep:
        dataclass = dataclass.shallow()

    # It shouldbe the same from loading it from the repo:
    assert dataclass == context.repo.load(scoped_loc, deep=is_deep)

    # Finally create aggregate from the DC loaded above:
    if aggregate_class in cant_be_shallow_aggs and not is_deep:
        with pytest.raises(AssertionError):
            aggregate = aggregate_class(dataclass, context.with_loc(
                scoped_loc.popped()), is_deep=is_deep)
        return  # skip the rest of the test when class can't be shallow
    else:
        aggregate = aggregate_class(dataclass, context.with_loc(
            scoped_loc.popped()), is_deep=is_deep)

    # Its data should stay intact right after construction:
    assert aggregate._dataclass == dataclass
    assert isinstance(aggregate, aggregate_class)

    # Save aggregate to repo, shallowly:
    aggregate.save(deep=False)

    # try to load its dataclass back to ensure consistency
    assert aggregate._dataclass == context.repo.load(scoped_loc, deep=is_deep)



import json

import numpy as np
import pytest
from neuraxle.base import Identity
from neuraxle.hyperparams.distributions import RandInt, Uniform
from neuraxle.hyperparams.space import (HyperparameterSamples,
                                        HyperparameterSpace)
from neuraxle.metaopt.auto_ml import AutoML, RandomSearchSampler
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.validation import KFoldCrossValidationSplitter
from neuraxle.pipeline import Pipeline
from neuraxle.steps.data import DataShuffler
from neuraxle.steps.flow import TrainOnlyWrapper
from neuraxle.steps.sklearn import SKLearnWrapper
from sklearn.decomposition import PCA
from sklearn.ensemble import BaggingRegressor, GradientBoostingRegressor
from sklearn.linear_model import LinearRegression, SGDClassifier, SGDRegressor
from sklearn.metrics import median_absolute_error


def test_sklearn_wrapper_with_an_invalid_step():
    with pytest.raises(ValueError):
        SKLearnWrapper(Identity())


def test_sklearn_wrapper_fit_transform_with_predict():
    p = SKLearnWrapper(LinearRegression())
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1)
    expected_outputs = np.expand_dims(np.array(list(range(10, 20))), axis=-1)

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(outputs, expected_outputs)


def test_sklearn_wrapper_transform_with_predict():
    p = SKLearnWrapper(LinearRegression())
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1)
    expected_outputs = np.expand_dims(np.array(list(range(10, 20))), axis=-1)

    p = p.fit(data_inputs, expected_outputs)
    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, expected_outputs)


def test_sklearn_wrapper_fit_transform_with_transform():
    n_components = 2
    p = SKLearnWrapper(PCA(n_components=n_components))
    dim1 = 10
    dim2 = 10
    data_inputs, expected_outputs = _create_data_source((dim1, dim2))

    p, outputs = p.fit_transform(data_inputs, expected_outputs)

    assert outputs.shape == (dim1, n_components)


def test_sklearn_wrapper_transform_partial_fit_with_predict():
    model = SKLearnWrapper(SGDRegressor(learning_rate='adaptive', eta0=0.05), use_partial_fit=True)
    p = Pipeline([TrainOnlyWrapper(DataShuffler()), model])
    data_inputs = np.expand_dims(np.array(list(range(10))), axis=-1) / 10
    expected_outputs = np.ravel(np.expand_dims(np.array(list(range(10, 20))), axis=-1)) / 10

    for _ in range(30):
        p = p.fit(data_inputs, expected_outputs)
    outputs = p.predict(data_inputs)

    assert all([np.isclose(a, b, atol=0.1) for a, b in zip(expected_outputs, outputs)])


def test_sklearn_wrapper_transform_partial_fit_classifier():
    data_inputs = np.array([[0, 1], [0, 0], [3, -2], [-1, 1], [-2, 1], [2, 0], [2, -1], [4, -2], [-3, 1], [-1, 0]])
    expected_outputs = np.ravel(np.expand_dims(data_inputs[:, 0] + 2 * data_inputs[:, 1] + 1, axis=-1))
    data_inputs = data_inputs / (4 + 1)
    classes = np.array([0, 1, 2, 3])
    model = SKLearnWrapper(
        SGDClassifier(learning_rate='adaptive', eta0=0.05),
        use_partial_fit=True,
        partial_fit_kwargs={'classes': classes}
    )
    p = Pipeline([TrainOnlyWrapper(DataShuffler()), model])

    for _ in range(30):
        p = p.fit(data_inputs, expected_outputs)
    outputs = p.predict(data_inputs)

    assert outputs.shape == (10,)
    assert len(set(outputs) - set(classes)) == 0


def test_sklearn_wrapper_set_hyperparams():
    p = SKLearnWrapper(PCA())
    p.set_hyperparams(HyperparameterSamples({
        'n_components': 2
    }))

    assert p.wrapped_sklearn_predictor.n_components == 2


def test_sklearn_wrapper_update_hyperparams():
    p = SKLearnWrapper(PCA())
    p.set_hyperparams(HyperparameterSamples({
        'n_components': 2,
        'svd_solver': 'full'
    }))
    p.update_hyperparams(HyperparameterSamples({
        'n_components': 4
    }))

    assert p.wrapped_sklearn_predictor.n_components == 4
    assert p.wrapped_sklearn_predictor.svd_solver == 'full'


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs


# With AutoML loop

def _test_within_auto_ml_loop(tmpdir, pipeline):
    X_train = np.random.random((25, 50)).astype(np.float32)
    Y_train = np.random.random((25,)).astype(np.float32)

    validation_splitter = KFoldCrossValidationSplitter(3)
    scoring_callback = ScoringCallback(
        median_absolute_error, higher_score_is_better=False)

    auto_ml = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=RandomSearchSampler(),
        validation_splitter=validation_splitter,
        scoring_callback=scoring_callback,
        n_trials=2,
        epochs=1,
        hyperparams_repository=HyperparamsOnDiskRepository(cache_folder=tmpdir),
        refit_best_trial=True,
        continue_loop_on_error=False)

    auto_ml.fit(X_train, Y_train)


@pytest.mark.skip(reason="AutoML loop refactor")
def test_automl_sklearn(tmpdir):
    grad_boost = SKLearnWrapper(GradientBoostingRegressor())
    _test_within_auto_ml_loop(tmpdir, grad_boost)


@pytest.mark.skip(reason="AutoML loop refactor")
def test_automl_sklearn_model_with_base_estimator(tmpdir):
    grad_boost = GradientBoostingRegressor()
    bagged_regressor = BaggingRegressor(
        grad_boost, random_state=5, n_jobs=-1)

    wrapped_bagged_regressor = SKLearnWrapper(
        bagged_regressor,
        HyperparameterSpace({
            "n_estimators": RandInt(10, 100),
            "max_features": Uniform(0.6, 1.0)}),
        #  return_all_sklearn_default_params_on_get=True
    )
    _test_within_auto_ml_loop(tmpdir, wrapped_bagged_regressor)



import numpy as np
import pytest

from neuraxle.base import ExecutionMode
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import TrainOrTestOnlyWrapper, TestOnlyWrapper
from neuraxle.steps.misc import TapeCallbackFunction, FitTransformCallbackStep
from testing.steps.neuraxle_test_case import NeuraxleTestCase

DATA_INPUTS = np.array([1])
EXPECTED_OUTPUTS = np.array([1])

tape_transform = TapeCallbackFunction()
tape_fit = TapeCallbackFunction()


@pytest.mark.parametrize('test_case', [
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        expected_callbacks_data=[[DATA_INPUTS]],
        data_inputs=DATA_INPUTS,
        expected_processed_outputs=DATA_INPUTS * 2,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[DATA_INPUTS], [(DATA_INPUTS, EXPECTED_OUTPUTS)]],
        expected_processed_outputs=DATA_INPUTS * 2,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], [(DATA_INPUTS, EXPECTED_OUTPUTS)]],
        execution_mode=ExecutionMode.FIT
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        expected_callbacks_data=[[], []],
        data_inputs=DATA_INPUTS,
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], []],
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], []],
        execution_mode=ExecutionMode.FIT
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        expected_callbacks_data=[[DATA_INPUTS]],
        data_inputs=DATA_INPUTS,
        expected_processed_outputs=DATA_INPUTS * 2,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[DATA_INPUTS], [(DATA_INPUTS, EXPECTED_OUTPUTS)]],
        expected_processed_outputs=DATA_INPUTS * 2,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], [(DATA_INPUTS, EXPECTED_OUTPUTS)]],
        execution_mode=ExecutionMode.FIT
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        expected_callbacks_data=[[], []],
        data_inputs=DATA_INPUTS,
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], []],
        expected_processed_outputs=DATA_INPUTS,
        execution_mode=ExecutionMode.FIT_TRANSFORM
    ),
    NeuraxleTestCase(
        pipeline=TrainOrTestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2), is_train_only=False),
        more_arguments={'set_train': True},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], []],
        execution_mode=ExecutionMode.FIT
    ),
    NeuraxleTestCase(
        pipeline=TestOnlyWrapper(FitTransformCallbackStep(tape_transform, tape_fit, transform_function=lambda di: di * 2)),
        more_arguments={'set_train': False},
        callbacks=[tape_transform, tape_fit],
        data_inputs=DATA_INPUTS,
        expected_outputs=EXPECTED_OUTPUTS,
        expected_callbacks_data=[[], [(DATA_INPUTS, EXPECTED_OUTPUTS)]],
        execution_mode=ExecutionMode.FIT
    )
])
def test_train_only_wrapper(test_case: NeuraxleTestCase):
    test_case.pipeline.set_train(test_case.more_arguments['set_train'])

    processed_outputs = test_case.execute()

    test_case.assert_expected_processed_outputs(processed_outputs)
    test_case.assert_callback_data_is_as_expected()



from unittest import TestCase

import numpy as np
import pytest
from neuraxle.base import AssertExpectedOutputIsNone, BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import ExecutionPhase, HandleOnlyMixin, NonFittableMixin
from neuraxle.data_container import DataContainer as DACT
from neuraxle.pipeline import Pipeline


class SomeAssertionStep(NonFittableMixin, HandleOnlyMixin, BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        HandleOnlyMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        _, data_inputs, expected_outputs = data_container.tolist().unpack()
        if expected_outputs is not None:
            self._assert_equals(data_inputs, expected_outputs, "Assertion failed", context)
        return data_inputs


class TestAssertionMethodInSteps(TestCase):

    def test_assertion_step_logs_and_raises_with_pipeline(self):
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        dact = DACT(data_inputs, None, expected_outputs)
        p = Pipeline([SomeAssertionStep()])

        with self.assertLogs() as captured:
            with pytest.raises(AssertionError):
                p.handle_fit_transform(dact, context=CX())

            self.assertIn("Assertion failed", captured.output[0])

    def test_assertion_step_just_logs_with_pipeline_in_prod(self):
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        dact = DACT(data_inputs, None, expected_outputs)
        p = Pipeline([SomeAssertionStep()])
        context = CX(execution_phase=ExecutionPhase.PROD)
        try:
            p = p.handle_fit(dact, context=context)
        except AssertionError:
            pass

        with self.assertLogs() as captured:
            p.handle_predict(dact, context=context)

            # assert that the log still at least contains the expected message:
            self.assertIn("Assertion failed", captured.output[0])


def test_expectedoutputnull_raise_exception_when_notnull(tmpdir):
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 2

    p = Pipeline([AssertExpectedOutputIsNone()])

    with pytest.raises(AssertionError) as error_info:
        p.fit_transform(data_inputs, expected_outputs)


def test_expectedoutputnull_is_fine_when_null(tmpdir):
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = None

    p = Pipeline([AssertExpectedOutputIsNone()])
    p.fit_transform(data_inputs, expected_outputs)



import pytest
import numpy as np
from sklearn.metrics import mean_squared_error
from neuraxle.base import ExecutionContext as CX

from neuraxle.hyperparams.distributions import RandInt
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.callbacks import MetricCallback
from neuraxle.metaopt.validation import KFoldCrossValidationSplitter, ValidationSplitter, RandomSearchSampler
from neuraxle.metaopt.auto_ml import AutoML
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN
from neuraxle.data_container import DataContainer as DACT


def test_automl_sequence_splitter(tmpdir):
    # Setting seed for better reproducibility
    np.random.seed(68)

    # Given
    data_inputs = np.array(range(100))
    expected_outputs = np.array(range(100, 200))

    hyperparameter_space = HyperparameterSpace({
        'multiplication_1__multiply_by': RandInt(1, 3),
        'multiplication_2__multiply_by': RandInt(1, 3),
        'multiplication_3__multiply_by': RandInt(1, 3),
    })

    pipeline = Pipeline([
        ('multiplication_1', MultiplyByN()),
        ('multiplication_2', MultiplyByN()),
        ('multiplication_3', MultiplyByN())
    ]).set_hyperparams_space(hyperparameter_space)

    auto_ml = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=RandomSearchSampler(),
        validation_splitter=KFoldCrossValidationSplitter(k_fold=4),
        callbacks=[MetricCallback("MSE", mean_squared_error, False)],
    )

    # When
    auto_ml = auto_ml.handle_fit(
        DACT(data_inputs=data_inputs, expected_outputs=expected_outputs), CX(tmpdir))
    predicted_outputs = auto_ml.transform(data_inputs)

    # Then
    actual_mse = ((predicted_outputs - expected_outputs) ** 2).mean()
    assert actual_mse < 20000


def test_automl_validation_splitter(tmpdir):
    # Setting seed for reproducibility
    np.random.seed(75)
    # Given
    data_inputs = np.array(range(1000, 1020))
    expected_outputs = np.array(range(2020, 2040))

    hyperparameter_space = HyperparameterSpace({
        'multiplication_1__multiply_by': RandInt(1, 3),
        'multiplication_2__multiply_by': RandInt(1, 3),
        'multiplication_3__multiply_by': RandInt(1, 3),
    })

    pipeline = Pipeline([
        ('multiplication_1', MultiplyByN()),
        ('multiplication_2', MultiplyByN()),
        ('multiplication_3', MultiplyByN())
    ]).set_hyperparams_space(hyperparameter_space)

    hp_search = AutoML(
        pipeline=pipeline,
        validation_splitter=ValidationSplitter(validation_size=0.2),
        scoring_callback=MetricCallback("MSE", mean_squared_error, False),
        n_trials=18,
    ).with_context(CX(tmpdir))

    # When
    mse_before = ((data_inputs - expected_outputs) ** 2).mean()
    hp_search = hp_search.fit(data_inputs, expected_outputs)
    predicted_outputs = hp_search.transform(data_inputs)

    # Then
    actual_mse = ((predicted_outputs - expected_outputs) ** 2).mean()
    assert actual_mse < mse_before



"""
Parallel processing in Neuraxle
===================================================================

This demonstrates how to stream data in parallel in a Neuraxle pipeline.
The pipeline steps' parallelism here will be obvious.

The pipeline has two steps:
1. Preprocessing: the step that process the data simply sleeps.
2. Model: the model simply multiplies the data by two.

This can be used with scikit-learn as well to transform things in parallel,
and any other library such as tensorflow.

Pipelines benchmarked:
1. We first use a classical pipeline and evaluate the time.
2. Then we use a minibatched pipeline and we evaluate the time.
3. Then we use a parallel pipeline and we evaluate the time.

We expect the parallel pipeline to be faster due to having more workers
in parallel, as well as starting the model's transformations at the same
time that other batches are being preprocessed, using queues.


..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import time
import numpy as np
from neuraxle.base import ExecutionContext as CX

from neuraxle.distributed.streaming import SequentialQueuedPipeline
from neuraxle.pipeline import BasePipeline, Pipeline, MiniBatchSequentialPipeline
from neuraxle.steps.loop import ForEach
from neuraxle.steps.misc import Sleep
from neuraxle.steps.numpy import MultiplyByN


def eval_run_time(pipeline: BasePipeline):
    pipeline.setup(CX())
    a = time.time()
    output = pipeline.transform(list(range(100)))
    b = time.time()
    seconds = b - a
    return seconds, output


def main():
    """
    The task is to sleep 0.02 seconds for each data input and then multiply by 2.
    """
    sleep_time = 0.02
    preprocessing_and_model_steps = [ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)]

    # Classical pipeline - all at once with one big batch:
    p = Pipeline(preprocessing_and_model_steps)
    time_vanilla_pipeline, output_classical = eval_run_time(p)
    print(f"Classical 'Pipeline' execution time: {time_vanilla_pipeline} seconds.")

    # Classical minibatch pipeline - minibatch size 25:
    p = MiniBatchSequentialPipeline(preprocessing_and_model_steps,
                                    batch_size=25)
    time_minibatch_pipeline, output_minibatch = eval_run_time(p)
    print(f"Minibatched 'MiniBatchSequentialPipeline' execution time: {time_minibatch_pipeline} seconds.")

    # Parallel pipeline - minibatch size 25 with 4 parallel workers per step that
    # have a max queue size of 10 batches between preprocessing and the model:
    p = SequentialQueuedPipeline(preprocessing_and_model_steps,
                                 n_workers_per_step=4, max_queue_size=10, batch_size=25)
    time_parallel_pipeline, output_parallel = eval_run_time(p)
    print(f"Parallel 'SequentialQueuedPipeline' execution time: {time_parallel_pipeline} seconds.")

    assert time_parallel_pipeline < time_minibatch_pipeline, str((time_parallel_pipeline, time_vanilla_pipeline))
    assert np.array_equal(output_classical, output_minibatch)
    assert np.array_equal(output_classical, output_parallel)


if __name__ == '__main__':
    main()



"""
Neuraxle's AutoML Classes
====================================
Classes used to build any Automatic Machine Learning pipelines.
Hyperparameter selection strategies are used to optimize the hyperparameters of given pipelines.

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import gc
from copy import copy
from typing import ContextManager, Iterator, List, Optional, Tuple

from neuraxle.base import (CX, BaseService, BaseStep, BaseStepT,
                           ExecutionContext, ForceHandleMixin,
                           TruncableService, _HasChildrenMixin)
from neuraxle.data_container import IDT
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.callbacks import (ARG_Y_EXPECTED, ARG_Y_PREDICTD,
                                        BaseCallback, CallbackList,
                                        ScoringCallback)
from neuraxle.metaopt.data.aggregates import (Client, Project, Root, Round,
                                              Trial, TrialSplit)
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT,
                                           AutoMLContext,
                                           HyperparamsRepository)
from neuraxle.metaopt.validation import (BaseHyperparameterOptimizer,
                                         BaseValidationSplitter,
                                         GridExplorationSampler,
                                         RandomSearchSampler,
                                         ValidationSplitter)


class Trainer(BaseService):
    """
    Class used to train a pipeline using various data splits and callbacks for evaluation purposes.
    """
    # TODO: add this `with_val_set` method that would change splitter to
    # PresetValidationSetSplitter(self, val) and override.

    def __init__(
            self,
            validation_splitter: 'BaseValidationSplitter',
            callbacks: List[BaseCallback] = None,
            n_epochs: int = 1,
    ):
        BaseService.__init__(self)
        self.validation_splitter: BaseValidationSplitter = validation_splitter
        self.callbacks: CallbackList = CallbackList(callbacks) if callbacks is not None else []
        self.n_epochs = n_epochs

    def train(
        self,
        pipeline: BaseStep,
        dact: DACT,
        trial_scope: Trial,
        return_trained_pipelines: bool = False
    ) -> Optional[List[BaseStep]]:
        """
        Train pipeline using the validation splitter.
        Track training, and validation metrics for each epoch.
        Note: the present method is just a shortcut to using the `execute_trial` method with less boilerplate code needed. Refer to `execute_trial` for full flexibility
        """
        trained_pipelines: List[BaseStep] = []

        splits: List[Tuple[DACT, DACT]] = self.validation_splitter.split_dact(
            dact, context=trial_scope.context)

        for train_dact, val_dact in splits:
            with trial_scope.new_validation_split() as trial_split_scope:
                p = self.train_split(pipeline, train_dact, val_dact, trial_split_scope)
                if return_trained_pipelines:
                    trained_pipelines.append(p)
                else:
                    del p
                    gc.collect()

        return trained_pipelines

    def train_split(
        self,
        pipeline: BaseStep,
        train_dact: DACT,
        val_dact: Optional[DACT],
        trial_split_scope: TrialSplit
    ) -> BaseStep:
        """
        Train a pipeline split. You probably want to use `self.train` instead, to use the validation splitter.
        If validation DACT is None, the evaluation metrics will not save validation results.
        """
        trial_split_scope: TrialSplit = trial_split_scope.with_n_epochs(self.n_epochs)
        p: BaseStep = pipeline.copy(trial_split_scope.context, deep=True)
        p.set_hyperparams(trial_split_scope.get_hyperparams())

        for _ in range(self.n_epochs):
            e = trial_split_scope.next_epoch()

            # Fit train
            p = p.set_train(True)
            p = p.handle_fit(
                train_dact.copy(),
                trial_split_scope.context.train())

            # Predict train & val
            p = p.set_train(False)
            eval_dact_train = p.handle_predict(
                train_dact.without_eo(),
                trial_split_scope.context.validation())
            eval_dact_train: DACT[IDT, ARG_Y_PREDICTD, ARG_Y_EXPECTED] = eval_dact_train.with_eo(train_dact.eo)

            if val_dact is not None:
                eval_dact_valid = p.handle_predict(
                    val_dact.without_eo(),
                    trial_split_scope.context.validation())
                eval_dact_valid: DACT[IDT, ARG_Y_PREDICTD, ARG_Y_EXPECTED] = eval_dact_valid.with_eo(val_dact.eo)
            else:
                eval_dact_valid = None

            # Log metrics / evaluate
            if self.callbacks.call(
                trial_split_scope,
                eval_dact_train,
                eval_dact_valid,
                e == self.n_epochs
            ):
                break  # Saves stats using the '__exit__' method of managed scoped aggregates.

        return p

    def refit(
        self,
        pipeline: BaseStep,
        dact: DACT,
        trial_scope: Trial,
    ) -> BaseStep:
        """
        Refit the pipeline on the whole dataset (without any validation technique).

        :return: fitted pipeline
        """

        with trial_scope.retrain_split() as trial_split_scope:
            trial_split_scope: TrialSplit = trial_split_scope

            return self.train_split(
                pipeline,
                dact,
                None,
                trial_split_scope
            )


class BaseControllerLoop(TruncableService):

    def __init__(
        self,
        trainer: Trainer,
        n_trials: int,
        hp_optimizer: BaseHyperparameterOptimizer = None,
        continue_loop_on_error: bool = True
    ):
        if hp_optimizer is None:
            hp_optimizer = RandomSearchSampler()

        TruncableService.__init__(self, {
            Trainer: trainer,
            BaseHyperparameterOptimizer: hp_optimizer,
        })
        self.n_trials = n_trials
        self.continue_loop_on_error: bool = continue_loop_on_error

    @property
    def trainer(self) -> Trainer:
        return self.get_service(Trainer)

    def run(self, pipeline: BaseStep, dact: DACT, round_scope: Round):
        """
        Run the controller loop.

        :param context: execution context
        :return: the ID of the round that was executed (either created or continued from previous optimization).
        """

        # thread_safe_lock, context = context.thread_safe() ??

        hp_optimizer: BaseHyperparameterOptimizer = self[BaseHyperparameterOptimizer]
        hp_space: HyperparameterSpace = pipeline.get_hyperparams_space()

        round_scope: Round = round_scope.with_optimizer(hp_optimizer, hp_space)

        for managed_trial_scope in self.loop(round_scope):
            managed_trial_scope: Trial = managed_trial_scope  # typing helps
            # trial_scope.context.restore_lock(thread_safe_lock) ??
            self.trainer.train(pipeline, dact, managed_trial_scope)

    def loop(self, round_scope: Round) -> Iterator[Trial]:
        """
        Loop over all trials.

        :param dact: data container that is not yet splitted
        :param context: execution context
        :return:
        """
        for _ in range(self.n_trials):
            with self.next_trial(round_scope) as managed_trial_scope:
                managed_trial_scope: Trial = managed_trial_scope  # typing helps

                yield managed_trial_scope

    def next_trial(self, round_scope: Round) -> ContextManager[Trial]:
        """
        Get the next trial to be executed.

        :param round_scope: round scope
        :return: the next trial to be executed.
        """
        return round_scope.new_rvs_trial(self.continue_loop_on_error)  # TODO: parallelization with this method?

    def refit_best_trial(self, pipeline: BaseStep, dact: DACT, round_scope: Round) -> BaseStep:
        """
        Refit the pipeline on the whole dataset (without any validation technique).
        """
        with round_scope.refitting_best_trial() as managed_trial_scope:

            refitted: BaseStep = self.trainer.refit(
                pipeline, dact, managed_trial_scope)

        return refitted

    def for_refit_only(self) -> 'BaseControllerLoop':
        """
        Create a controller loop configured with zero iterations
        so as to only make the "refit_best_trial" possible.
        """
        self_copy = copy.copy(self)
        self_copy.n_trials = 0
        return self_copy


class DefaultLoop(BaseControllerLoop):

    def __init__(
        self,
        trainer: Trainer,
        n_trials: int,
        hp_optimizer: BaseHyperparameterOptimizer = None,
        continue_loop_on_error: bool = False,
        n_jobs: int = 1,
    ):
        super().__init__(
            trainer,
            n_trials,
            hp_optimizer=hp_optimizer,
            continue_loop_on_error=continue_loop_on_error
        )
        self.n_jobs = n_jobs


class ControlledAutoML(ForceHandleMixin, _HasChildrenMixin[BaseStepT], BaseStep):
    """
    A step to execute Automated Machine Learning (AutoML) algorithms. This step will
    automatically split the data into train and validation splits, and execute an
    hyperparameter optimization on the splits to find the best hyperparameters.

    The Controller Loop is useful to possibly split the execution into multiple
    threads, or even multiple machines.

    The Trainer is responsible for training the pipeline on the train and validation
    splits as splitted.

    The step with the chosen good hyperparameters will be refitted to the full
    unsplitted data if desired.
    """

    def __init__(
            self,
            pipeline: BaseStepT,
            loop: BaseControllerLoop,
            main_metric_name: str,
            repo: HyperparamsRepository = None,
            start_new_round: bool = True,
            refit_best_trial: bool = True,
            project_name: str = DEFAULT_PROJECT,
            client_name: str = DEFAULT_CLIENT,
    ):
        """
        .. note::
            Usage of a multiprocess-safe hyperparams repository is recommended,
            although it is, most of the time, not necessary.
            Beware of the behaviour of HyperparamsRepository's observers/subscribers.
            Context instances are not shared between trial but copied.
            So is the AutoML loop and the DACTs.

        :param pipeline: The pipeline, or BaseStep, which will be use by the AutoMLloop
        :param loop: The loop, or BaseControllerLoop, which will be used by the AutoML loop
        :param flow: The flow, or Flow, which will be used by the AutoML loop
        :param refit_best_trial: A boolean indicating whether to perform, after a fit call, a refit on the best trial.
        """
        BaseStep.__init__(self)
        _HasChildrenMixin.__init__(self)
        ForceHandleMixin.__init__(self)

        self.pipeline: BaseStepT = pipeline
        self.loop: BaseControllerLoop = loop
        self.repo: HyperparamsRepository = repo

        self.main_metric_name: str = main_metric_name
        self.start_new_round: bool = start_new_round
        self.refit_best_trial: bool = refit_best_trial
        self.project_name: str = project_name
        self.client_name: str = client_name

        self.has_model_been_retrained: bool = False

    def get_children(self) -> List[BaseStep]:
        return [self.pipeline]

    @property
    def wrapped(self) -> BaseStep:
        return self.pipeline

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        if not self.refit_best_trial:
            raise ValueError(
                "self.refit_best_trial must be True in this AutoML class to do the transform in 'fit_transform'.")

        self = self._fit_data_container(data_container, context)
        data_container = self._transform_data_container(data_container, context)
        return self, data_container

    def to_force_refit_best_trial(self) -> 'ControlledAutoML':
        self_copy = copy.copy(self)
        self_copy.refit_best_trial = True
        self_copy.has_model_been_retrained = False
        self_copy.loop = self_copy.loop.for_refit_only()
        return self_copy

    def _fit_data_container(self, data_container: DACT, context: CX) -> 'BaseStep':
        """
        Run Auto ML Loop.
        Find the best hyperparams using the hyperparameter optmizer.
        Evaluate the pipeline on each trial using a validation technique.

        :param data_container: data container to fit
        :param context: execution context

        :return: self
        """
        automl_context: AutoMLContext = self.get_automl_context(context)
        root: Root = Root.from_context(automl_context)

        with root.get_project(self.project_name) as ps:
            ps: Project = ps
            with ps.get_client(self.client_name) as cs:
                cs: Client = cs
                with cs.optim_round(self.start_new_round, self.main_metric_name) as rs:
                    rs: Round = rs

                    self.loop.run(self.pipeline, data_container, rs)

                    if self.refit_best_trial:
                        self.pipeline = self.loop.refit_best_trial(self.pipeline, data_container, rs)
                        self.has_model_been_retrained = True

        return self

    def get_automl_context(self, context: ExecutionContext) -> AutoMLContext:
        # if with_loc:
        #     return AutoMLContext.from_context(context, repo=self.repo).with_loc(self.project_name, self.client_name, ???)
        return AutoMLContext.from_context(context, repo=self.repo)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        if not self.has_model_been_retrained:
            raise ValueError(
                'self.refit_best_trial must be True in AutoML class '
                'to transform, and the AutoMl should have been fitted.')

        return self.pipeline.handle_transform(data_container, context)


class AutoML(ControlledAutoML):
    """
    This class provides a nice interface to easily use the
    ControlledAutoML class and the metaopt module in general.

    :param pipeline: pipeline to copy and use for training
    :param validation_splitter: validation splitter to use
    :param refit_best_trial: whether to refit the best model on the whole dataset after the optimization
    :param scoring_callback: main callback to use for scoring, that is deprecated
    :param hyperparams_optimizer: hyperparams optimizer to use
    :param hyperparams_repository: hyperparams repository to use
    :param n_trials: number of trials to run
    :param epochs: number of epochs to train the model for each val split
    :param callbacks: callbacks to use for training - there can be aditionnal metrics there
    :param n_jobs: number of jobs to use for parallelization, defaults is None for no parallelization
    :param continue_loop_on_error: whether to continue the main optimization loop on error or not
    :return: AutoML object ready to use with fit and transform.
    """

    def __init__(
        self,
        pipeline: BaseStep,
        validation_splitter: 'BaseValidationSplitter' = None,
        hyperparams_optimizer: BaseHyperparameterOptimizer = None,
        scoring_callback: ScoringCallback = None,
        callbacks: List[BaseCallback] = None,
        hyperparams_repository: HyperparamsRepository = None,
        n_trials: int = None,
        refit_best_trial: bool = True,
        epochs: int = 1,
        n_jobs=1,
        continue_loop_on_error=True
    ):
        # parse or guess args:
        validation_splitter = validation_splitter or ValidationSplitter(0.2)
        hyperparams_optimizer = hyperparams_optimizer or GridExplorationSampler()

        callbacks = list(callbacks) if callbacks is not None else []
        if scoring_callback is not None:
            callbacks = [scoring_callback] + callbacks
        if len(callbacks) == 0:
            raise ValueError("At least one callback must be provided.")

        if n_trials is None or n_trials < 1:
            n_trials: int = GridExplorationSampler.estimate_ideal_n_trials(pipeline.get_hyperparams_space())

        # init subservices:
        trainer = Trainer(
            callbacks=callbacks,
            validation_splitter=validation_splitter,
            n_epochs=epochs,
        )
        controller_loop = DefaultLoop(
            trainer=trainer,
            n_trials=n_trials,
            n_jobs=n_jobs,
            hp_optimizer=hyperparams_optimizer,
            continue_loop_on_error=continue_loop_on_error,
        )

        # init base class:
        ControlledAutoML.__init__(
            self,
            pipeline=pipeline,
            loop=controller_loop,
            repo=hyperparams_repository,
            main_metric_name=callbacks[0].name,
            start_new_round=True,
            refit_best_trial=refit_best_trial,
        )



from neuraxle.base import BaseStep, TruncableSteps, MetaStep, BaseTransformer
from neuraxle.hyperparams.distributions import LogUniform, Quantized, RandInt, Boolean
from neuraxle.hyperparams.space import HyperparameterSpace, HyperparameterSamples

HYPERPARAMETERS_SPACE = HyperparameterSpace({
    'learning_rate': LogUniform(0.0001, 0.1),
    'l2_weight_reg': LogUniform(0.0001, 0.1),
    'momentum': LogUniform(0.01, 1.0),
    'hidden_size': Quantized(LogUniform(16, 512)),
    'num_layers': RandInt(1, 4),
    'num_lstm_layers': RandInt(1, 2),
    'use_xavier_init': Boolean(),
    'use_max_pool_else_avg_pool': Boolean(),
    'dropout_drop_proba': LogUniform(0.3, 0.7)
})

HYPERPARAMETERS = HyperparameterSamples({
    'learning_rate': 0.1,
    'l2_weight_reg': 0.001,
    'hidden_size': 32,
    'num_layers': 3,
    'num_lstm_layers': 1,
    'use_xavier_init': True,
    'use_max_pool_else_avg_pool': True,
    'dropout_drop_proba': 0.5,
    'momentum': 0.1
})

AN_INPUT = "I am an input"
AN_EXPECTED_OUTPUT = "I am an expected output"


class SomeStep(BaseTransformer):
    def __init__(self, hyperparams_space: HyperparameterSpace = None, output=AN_EXPECTED_OUTPUT):
        super().__init__(hyperparams=None, hyperparams_space=hyperparams_space)
        self.output = output

    def transform(self, data_inputs):
        return [self.output] * len(data_inputs)


class SomeStepWithHyperparams(BaseStep):
    def __init__(self):
        super().__init__(
            hyperparams=HYPERPARAMETERS,
            hyperparams_space=HYPERPARAMETERS_SPACE,
            name="MockStep"
        )

    def transform(self, data_inputs):
        pass

    def fit(self, data_inputs, expected_outputs=None):
        pass


class SomeMetaStepWithHyperparams(MetaStep):
    def __init__(self):
        MetaStep.__init__(self, wrapped=SomeStepWithHyperparams())

    def transform(self, data_inputs):
        pass

    def fit(self, data_inputs, expected_outputs=None):
        pass


class SomeTruncableStep(TruncableSteps):
    def __init__(self):
        TruncableSteps.__init__(self,
                                hyperparams=HYPERPARAMETERS,
                                hyperparams_space=HYPERPARAMETERS_SPACE,
                                steps_as_tuple=(SomeStepWithHyperparams(), SomeStepWithHyperparams())
                                )

    def transform(self, data_inputs):
        pass

    def fit(self, data_inputs, expected_outputs=None):
        pass


class SomeSplitStep(BaseStep):
    def transform(self, data_inputs):
        pass



"""
Create Pipeline Steps in Neuraxle that doesn't fit or transform
================================================================

If a pipeline step doesn't need to be fitted and only transforms data (e.g.: taking the logarithm of the data),
then you can inherit from the NonFittableMixin as demonstrated here, which will override the fit method properly
for you. You can also use a NonTransformableMixin if your step doesn't transform anything, which is rarer. If your step
simply just does nothing to the data, then you could even use the Identity class of Neuraxle, which is simply a class
that inherits from both the NonFittableMixin, the NonTransformableMixin, and BaseStep.

Mixins are an old Object Oriented Programming (OOP) design pattern that resurfaces when designing
Machine Learning Pipelines. Those are add-ons to classes to implement some methods in some specific ways already.
A mixin doesn't inherit from BaseStep itself, because we can combine many of them in one class. However, a mixin must
suppose that the object that inherits from the mixin also inherits from it's base class. Here, our base class is the
BaseStep class.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of tche License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import numpy as np

from neuraxle.base import NonTransformableMixin, Identity, BaseStep, NonFittableMixin
from neuraxle.pipeline import Pipeline


class NonFittableStep(NonFittableMixin, BaseStep):
    """
    Fit method is automatically implemented as changing nothing.
    Please make your steps inherit from NonFittableMixin, when they don't need any fitting.
    Also, make sure that BaseStep is the last step you inherit from.
    Note that we could also define the inverse_transform method in the present object.
    """
    def __init__(self):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

    def transform(self, data_inputs):
        # insert your transform code here
        print("NonFittableStep: I transformed.")
        return data_inputs


class NonTransformableStep(NonTransformableMixin, BaseStep):
    """
    Transform method is automatically implemented as returning data inputs as it is.
    Please make your steps inherit from NonTransformableMixin, when they don't need any transformations.
    Also, make sure that BaseStep is the last step you inherit from.
    """
    def __init__(self):
        BaseStep.__init__(self)
        NonTransformableMixin.__init__(self)

    def fit(self, data_inputs, expected_outputs=None) -> 'NonTransformableStep':
        # insert your fit code here
        print("NonTransformableStep: I fitted.")
        return self


def main():
    p = Pipeline([
        NonFittableStep(),
        NonTransformableStep(),
        Identity()  # Note: Identity does nothing: it inherits from both NonFittableMixin and NonTransformableMixin.
    ])

    some_data = np.array([0, 1])
    p = p.fit(some_data)
    # Out:
    #     NonFittableStep: I transformed.
    #     NonTransformableStep: I fitted.

    out = p.transform(some_data)
    # Out:
    #     NonFittableStep: I transformed.

    assert np.array_equal(out, some_data)
    # Data is unchanged as we've done nothing in the only transform.


if __name__ == "__main__":
    main()



"""
Usage of AutoML loop, and hyperparams with sklearn models.
=============================================================

This demonstrates how you can build an AutoML loop that finds the best possible sklearn classifier.
It also shows you how to add hyperparams to sklearn steps using SKLearnWrapper.
This example has been derived and simplified from the following repository: https://github.com/Neuraxio/Kata-Clean-Machine-Learning-From-Dirty-Code
Here, 2D data is fitted, whereas in the original example 3D (sequential / time series) data is preprocessed and then fitted with the same models.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import shutil

from neuraxle.base import ExecutionContext as CX
from neuraxle.hyperparams.distributions import (Boolean, Choice, LogUniform,
                                                RandInt)
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import (AutoML, RandomSearchSampler,
                                      ValidationSplitter)
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.pipeline import Pipeline
from neuraxle.steps.flow import ChooseOneStepOf
from neuraxle.steps.numpy import NumpyRavel
from neuraxle.steps.output_handlers import OutputTransformerWrapper
from neuraxle.steps.sklearn import SKLearnWrapper
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression, RidgeClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier, ExtraTreeClassifier


def main(tmpdir: str):
    # Define classification models, and hyperparams.

    decision_tree_classifier = SKLearnWrapper(
        DecisionTreeClassifier(),
        HyperparameterSpace({
            'criterion': Choice(['gini', 'entropy']),
            'splitter': Choice(['best', 'random']),
            'min_samples_leaf': RandInt(2, 5),
            'min_samples_split': RandInt(2, 4)
        }))

    extra_tree_classifier = SKLearnWrapper(
        ExtraTreeClassifier(),
        HyperparameterSpace({
            'criterion': Choice(['gini', 'entropy']),
            'splitter': Choice(['best', 'random']),
            'min_samples_leaf': RandInt(2, 5),
            'min_samples_split': RandInt(2, 4)
        }))

    ridge_classifier = Pipeline([OutputTransformerWrapper(NumpyRavel()), SKLearnWrapper(
        RidgeClassifier(),
        HyperparameterSpace({
            'alpha': Choice([0.0, 1.0, 10.0, 100.0]),
            'fit_intercept': Boolean(),
            'normalize': Boolean()
        }))
    ]).set_name('RidgeClassifier')

    logistic_regression = Pipeline([OutputTransformerWrapper(NumpyRavel()), SKLearnWrapper(
        LogisticRegression(),
        HyperparameterSpace({
            'C': LogUniform(0.01, 10.0),
            'fit_intercept': Boolean(),
            'penalty': Choice(['none', 'l2']),
            'max_iter': RandInt(20, 200)
        }))
    ]).set_name('LogisticRegression')

    random_forest_classifier = Pipeline([OutputTransformerWrapper(NumpyRavel()), SKLearnWrapper(
        RandomForestClassifier(),
        HyperparameterSpace({
            'n_estimators': RandInt(50, 600),
            'criterion': Choice(['gini', 'entropy']),
            'min_samples_leaf': RandInt(2, 5),
            'min_samples_split': RandInt(2, 4),
            'bootstrap': Boolean()
        }))
    ]).set_name('RandomForestClassifier')

    # Define a classification pipeline that lets the AutoML loop choose one of the classifier.
    # See also ChooseOneStepOf documentation: https://www.neuraxle.org/stable/api/neuraxle.steps.flow.html#neuraxle.steps.flow.ChooseOneStepOf

    pipeline = Pipeline([
        ChooseOneStepOf([
            decision_tree_classifier,
            extra_tree_classifier,
            ridge_classifier,
            logistic_regression,
            random_forest_classifier
        ])
    ])

    # Create the AutoML loop object.
    # See also AutoML documentation: https://www.neuraxle.org/stable/api/neuraxle.metaopt.auto_ml.html#neuraxle.metaopt.auto_ml.AutoML

    auto_ml = AutoML(
        pipeline=pipeline,
        hyperparams_optimizer=RandomSearchSampler(),
        validation_splitter=ValidationSplitter(validation_size=0.20),
        scoring_callback=ScoringCallback(accuracy_score, higher_score_is_better=True),
        n_trials=7,
        epochs=1,
        hyperparams_repository=HyperparamsOnDiskRepository(cache_folder=tmpdir),
        refit_best_trial=True,
        continue_loop_on_error=False
    )

    # Load data, and launch AutoML loop !

    X_train, y_train, X_test, y_test = generate_classification_data()
    auto_ml = auto_ml.fit(X_train, y_train)

    # Get the model from the best trial, and make predictions using predict, as per the `refit_best_trial=True` argument to AutoML.
    y_pred = auto_ml.predict(X_test)

    accuracy = accuracy_score(y_true=y_test, y_pred=y_pred)
    print("Test accuracy score:", accuracy)

    shutil.rmtree(tmpdir)


def generate_classification_data():
    data_inputs, expected_outputs = make_classification(
        n_samples=10000,
        n_repeated=0,
        n_classes=3,
        n_features=4,
        n_clusters_per_class=1,
        class_sep=1.5,
        flip_y=0,
        weights=[0.5, 0.5, 0.5]
    )

    X_train, X_test, y_train, y_test = train_test_split(
        data_inputs,
        expected_outputs,
        test_size=0.20
    )

    return X_train, y_train, X_test, y_test


if __name__ == '__main__':
    main(CX.get_new_cache_folder())



import numpy as np
import pytest
from neuraxle.base import BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import ForceHandleIdentity, ForceHandleMixin
from neuraxle.data_container import DataContainer as DACT
from neuraxle.pipeline import Pipeline


class BadForceHandleStep(ForceHandleMixin, BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        ForceHandleMixin.__init__(self)


def test_raises_exception_if_method_not_redefined(tmpdir):
    # Now that I think about it, this really just is a complicated way to test the self._ensure_method_overriden function.
    with pytest.raises(NotImplementedError) as exception_info:
        BadForceHandleStep()

    assert "Please define _fit_data_container" in exception_info.value.args[0]
    assert "in BadForceHandleStep" in exception_info.value.args[0]

    def _fit_data_container(self, data_container: DACT, context: CX):
        return self
    BadForceHandleStep._fit_data_container = _fit_data_container

    with pytest.raises(NotImplementedError) as exception_info:
        BadForceHandleStep()

    assert "Please define _fit_transform_data_container" in exception_info.value.args[0]
    assert "in BadForceHandleStep" in exception_info.value.args[0]

    def _fit_transform_data_container(self, data_container: DACT, context: CX):
        return self, data_container
    BadForceHandleStep._fit_transform_data_container = _fit_transform_data_container

    with pytest.raises(NotImplementedError) as exception_info:
        BadForceHandleStep()

    assert "Please define _transform_data_container" in exception_info.value.args[0]
    assert "in BadForceHandleStep" in exception_info.value.args[0]

    def _transform_data_container(self, data_container: DACT, context: CX):
        return data_container
    BadForceHandleStep._transform_data_container = _transform_data_container

    #Should not raise any error now.
    BadForceHandleStep()


def test_forcehandleidentity_does_not_crash(tmpdir):
    p = Pipeline([
        ForceHandleIdentity()
    ])
    data_inputs = np.array([0, 1, 2, 3])
    expected_outputs = data_inputs * 2
    p.fit(data_inputs, expected_outputs)
    p.fit_transform(data_inputs, expected_outputs)
    p.transform(data_inputs=data_inputs)



"""
Create Pipeline Steps that require implementing only handler methods
========================================================================================================================

If a pipeline step only needs to implement handler methods, then you can inherit from the
ForceHandleMixin as demonstrated here. Handler methods are useful when :

    - You need to change the shape of the data container passed to the following steps, or the wrapped steps.
    - You want to apply side effects based on the data container, and the execution context.
    - You want to change the pipeline execution flow based on the data container, and the execution context.

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
from typing import Tuple

import numpy as np
from neuraxle.base import DACT, BaseStep
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import ForceHandleMixin


class ForceHandleMixinStep(ForceHandleMixin, BaseStep):
    """
    Please make your steps inherit from ForceHandleMixin when they only implement handle_methods, but also
    when you want to make impossible the use of regular fit, transform, and fit_transform methods
    Also, make sure that BaseStep is the last step you inherit from.
    """

    def __init__(self):
        BaseStep.__init__(self)
        ForceHandleMixin.__init__(self)

    def _fit_data_container(self, data_container: DACT, context: CX) -> BaseStep:
        """
        Change the shape of the data container.
        and/or
        Apply any side effects based on the data container
        And/or
        Change the execution flow of the pipeline
        """
        context.logger.info("Handling the 'fit' with handler method!")
        return self

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Change the shape of the data container.
        and/or
        Apply any side effects based on the data container
        And/or
        Change the execution flow of the pipeline
        """
        context.logger.info("Handling the 'transform' with handler method!")
        return data_container

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[BaseStep, DACT]:
        """
        Change the shape of the data container.
        and/or
        Apply any side effects based on the data container
        And/or
        Change the execution flow of the pipeline
        """
        context.logger.info("Handling the 'fit_transform' with handler method!")
        return self, data_container


def main():
    p = ForceHandleMixinStep()
    data_inputs = np.array([0, 1])
    expected_outputs = np.array([0, 1])

    p = p.fit(data_inputs, expected_outputs)
    outputs = p.transform(data_inputs)


if __name__ == '__main__':
    main()



import time

import numpy as np
import pytest

from neuraxle.base import ExecutionContext as CX
from neuraxle.data_container import DACT, AbsentValuesNullObject
from neuraxle.distributed.streaming import SequentialQueuedPipeline, ParallelQueuedFeatureUnion, QueueJoiner
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import ForEach
from neuraxle.steps.misc import FitTransformCallbackStep, Sleep
from neuraxle.steps.numpy import MultiplyByN

EXPECTED_OUTPUTS = np.array(range(100)) * 2 * 2 * 2 * 2
EXPECTED_OUTPUTS_PARALLEL = np.array((np.array(range(100)) * 2).tolist() * 4)


def test_queued_pipeline_with_excluded_incomplete_batch():
    p = SequentialQueuedPipeline([
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2)
    ], batch_size=10, keep_incomplete_batch=False, n_workers_per_step=1, max_queue_size=5)

    outputs = p.transform(list(range(15)))

    assert np.array_equal(outputs, np.array(list(range(10))) * 2 * 2 * 2 * 2)


def test_queued_pipeline_with_included_incomplete_batch():
    p = SequentialQueuedPipeline(
        [
            MultiplyByN(2),
            MultiplyByN(2),
            MultiplyByN(2),
            MultiplyByN(2)
        ],
        batch_size=10,
        keep_incomplete_batch=True,
        default_value_data_inputs=AbsentValuesNullObject(),
        default_value_expected_outputs=AbsentValuesNullObject(),
        n_workers_per_step=1,
        max_queue_size=5
    )

    outputs = p.transform(list(range(15)))

    assert np.array_equal(outputs, np.array(list(range(15))) * 2 * 2 * 2 * 2)


def test_queued_pipeline_with_included_incomplete_batch_that_raises_an_exception():
    with pytest.raises(AttributeError):
        p = SequentialQueuedPipeline(
            [
                MultiplyByN(2),
                MultiplyByN(2),
                MultiplyByN(2),
                MultiplyByN(2)
            ],
            batch_size=10,
            keep_incomplete_batch=True,
            default_value_data_inputs=None,  # this will raise an exception in the worker
            default_value_expected_outputs=None,  # this will raise an exception in the worker
            n_workers_per_step=1,
            max_queue_size=5
        )
        p.transform(list(range(15)))


def test_queued_pipeline_with_step_with_process():
    p = SequentialQueuedPipeline([
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2)
    ], batch_size=10, n_workers_per_step=1, max_queue_size=5, use_processes=True)

    data_container = DACT(data_inputs=list(range(100)))
    context = CX()

    outputs = p.handle_transform(data_container, context)

    assert np.array_equal(outputs.data_inputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_step_with_threading():
    p = SequentialQueuedPipeline([
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2),
        MultiplyByN(2)
    ], batch_size=10, n_workers_per_step=1, max_queue_size=5, use_processes=False)

    data_container = DACT(data_inputs=list(range(100)))
    context = CX()

    outputs = p.handle_transform(data_container, context)

    assert np.array_equal(outputs.data_inputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_step_name_step():
    p = SequentialQueuedPipeline([
        ('1', MultiplyByN(2)),
        ('2', MultiplyByN(2)),
        ('3', MultiplyByN(2)),
        ('4', MultiplyByN(2))
    ], batch_size=10, n_workers_per_step=1, max_queue_size=5)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_n_workers_step():
    p = SequentialQueuedPipeline([
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2))
    ], batch_size=10, max_queue_size=5)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_wrapped_queued_pipeline_with_n_workers_step():
    p = Pipeline([SequentialQueuedPipeline([
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2)),
        (1, MultiplyByN(2))
    ], batch_size=10, max_queue_size=5)])

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_step_name_n_worker_max_queue_size():
    p = SequentialQueuedPipeline([
        ('1', 1, 5, MultiplyByN(2)),
        ('2', 1, 5, MultiplyByN(2)),
        ('3', 1, 5, MultiplyByN(2)),
        ('4', 1, 5, MultiplyByN(2))
    ], batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_step_name_n_worker_with_step_name_n_workers_and_default_max_queue_size():
    p = SequentialQueuedPipeline([
        ('1', 1, MultiplyByN(2)),
        ('2', 1, MultiplyByN(2)),
        ('3', 1, MultiplyByN(2)),
        ('4', 1, MultiplyByN(2))
    ], max_queue_size=10, batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_queued_pipeline_with_step_name_n_worker_with_default_n_workers_and_default_max_queue_size():
    p = SequentialQueuedPipeline([
        ('1', MultiplyByN(2)),
        ('2', MultiplyByN(2)),
        ('3', MultiplyByN(2)),
        ('4', MultiplyByN(2))
    ], n_workers_per_step=1, max_queue_size=10, batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_parallel_queued_pipeline_with_step_name_n_worker_max_queue_size():
    p = ParallelQueuedFeatureUnion([
        ('1', 1, 5, MultiplyByN(2)),
        ('2', 1, 5, MultiplyByN(2)),
        ('3', 1, 5, MultiplyByN(2)),
        ('4', 1, 5, MultiplyByN(2))
    ], batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS_PARALLEL)


def test_parallel_queued_threads_do_parallelize_sleep_correctly(tmpdir):
    sleep_time = 0.01
    p = SequentialQueuedPipeline([
        ('1', 2, 10, Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)])),
        ('2', 2, 10, Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)])),
        ('3', 2, 10, Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)])),
        ('4', 2, 10, Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)]))
    ], batch_size=10, use_processes=False, use_savers=False).with_context(CX(tmpdir))

    a = time.time()
    outputs_streaming = p.transform(list(range(100)))
    b = time.time()
    time_queued_pipeline = b - a

    p = Pipeline([
        Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)]),
        Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)]),
        Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)]),
        Pipeline([ForEach(Sleep(sleep_time=sleep_time)), MultiplyByN(2)])
    ])

    a = time.time()
    outputs_vanilla = p.transform(list(range(100)))
    b = time.time()
    time_vanilla_pipeline = b - a

    assert time_queued_pipeline < time_vanilla_pipeline
    assert np.array_equal(outputs_streaming, outputs_vanilla)


def test_parallel_queued_pipeline_with_step_name_n_worker_additional_arguments_max_queue_size():
    n_workers = 4
    worker_arguments = [('hyperparams', HyperparameterSamples({'multiply_by': 2})) for _ in range(n_workers)]
    p = ParallelQueuedFeatureUnion([
        ('1', n_workers, worker_arguments, 5, MultiplyByN()),
    ], batch_size=10)

    outputs = p.transform(list(range(100)))

    expected = np.array(list(range(0, 200, 2)))
    assert np.array_equal(outputs, expected)


def test_parallel_queued_pipeline_with_step_name_n_worker_additional_arguments():
    n_workers = 4
    worker_arguments = [('hyperparams', HyperparameterSamples({'multiply_by': 2})) for _ in range(n_workers)]
    p = ParallelQueuedFeatureUnion([
        ('1', n_workers, worker_arguments, MultiplyByN()),
    ], batch_size=10, max_queue_size=5)

    outputs = p.transform(list(range(100)))

    expected = np.array(list(range(0, 200, 2)))
    assert np.array_equal(outputs, expected)


def test_parallel_queued_pipeline_with_step_name_n_worker_with_step_name_n_workers_and_default_max_queue_size():
    p = ParallelQueuedFeatureUnion([
        ('1', 1, MultiplyByN(2)),
        ('2', 1, MultiplyByN(2)),
        ('3', 1, MultiplyByN(2)),
        ('4', 1, MultiplyByN(2)),
    ], max_queue_size=10, batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS_PARALLEL)


def test_parallel_queued_pipeline_with_step_name_n_worker_with_default_n_workers_and_default_max_queue_size():
    p = ParallelQueuedFeatureUnion([
        ('1', MultiplyByN(2)),
        ('2', MultiplyByN(2)),
        ('3', MultiplyByN(2)),
        ('4', MultiplyByN(2)),
    ], n_workers_per_step=1, max_queue_size=10, batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS_PARALLEL)


def test_parallel_queued_pipeline_step_name_n_worker_with_default_n_workers_and_default_max_queue_size():
    p = ParallelQueuedFeatureUnion([
        ('1', MultiplyByN(2)),
        ('2', MultiplyByN(2)),
        ('3', MultiplyByN(2)),
        ('4', MultiplyByN(2)),
    ], n_workers_per_step=1, max_queue_size=10, batch_size=10)

    outputs = p.transform(list(range(100)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS_PARALLEL)


@pytest.mark.parametrize("use_processes", [False, True])
@pytest.mark.parametrize("use_savers", [False, True])
def test_queued_pipeline_saving(tmpdir, use_processes, use_savers):
    # Given
    p = ParallelQueuedFeatureUnion([
        ('1', 4, 10, FitTransformCallbackStep()),
        ('2', 4, 10, FitTransformCallbackStep()),
        ('3', 4, 10, FitTransformCallbackStep()),
        ('4', 4, 10, FitTransformCallbackStep()),
    ], n_workers_per_step=4, max_queue_size=10, batch_size=10,
        use_processes=use_processes, use_savers=use_savers).with_context(CX(tmpdir))

    # When
    p, _ = p.fit_transform(list(range(200)), list(range(200)))
    p = p.wrapped  # clear execution context wrapper
    p.save(CX(tmpdir))
    p.apply('clear_callbacks')

    # Then

    assert len(p[0].wrapped.transform_callback_function.data) == 0
    assert len(p[0].wrapped.fit_callback_function.data) == 0
    assert len(p[1].wrapped.transform_callback_function.data) == 0
    assert len(p[1].wrapped.fit_callback_function.data) == 0
    assert len(p[2].wrapped.transform_callback_function.data) == 0
    assert len(p[2].wrapped.fit_callback_function.data) == 0
    assert len(p[3].wrapped.transform_callback_function.data) == 0
    assert len(p[3].wrapped.fit_callback_function.data) == 0

    p = p.load(CX(tmpdir))

    assert len(p[0].wrapped.transform_callback_function.data) == 20
    assert len(p[0].wrapped.fit_callback_function.data) == 20
    assert len(p[1].wrapped.transform_callback_function.data) == 20
    assert len(p[1].wrapped.fit_callback_function.data) == 20
    assert len(p[2].wrapped.transform_callback_function.data) == 20
    assert len(p[2].wrapped.fit_callback_function.data) == 20
    assert len(p[3].wrapped.transform_callback_function.data) == 20
    assert len(p[3].wrapped.fit_callback_function.data) == 20


@pytest.mark.parametrize("use_processes", [False, True])
def test_queued_pipeline_with_savers(tmpdir, use_processes: bool):
    # Given
    p = ParallelQueuedFeatureUnion([
        ('1', MultiplyByN(2)),
        ('2', MultiplyByN(2)),
        ('3', MultiplyByN(2)),
        ('4', MultiplyByN(2)),
    ], n_workers_per_step=1, max_queue_size=10, batch_size=10,
        use_savers=True, use_processes=use_processes
    ).with_context(CX(tmpdir))

    # When
    outputs = p.transform(list(range(100)))

    # Then
    assert np.array_equal(outputs, EXPECTED_OUTPUTS_PARALLEL)


class QueueJoinerForTest(QueueJoiner):
    def __init__(self, batch_size):
        super().__init__(batch_size)
        self.called_queue_joiner = False

    def join(self, original_data_container: DACT) -> DACT:
        self.called_queue_joiner = True
        super().join(original_data_container)


def test_sequential_queued_pipeline_should_fit_without_multiprocessing():
    batch_size = 10
    p = SequentialQueuedPipeline([
        (1, FitTransformCallbackStep()),
        (1, FitTransformCallbackStep()),
        (1, FitTransformCallbackStep()),
        (1, FitTransformCallbackStep())
    ], batch_size=batch_size, max_queue_size=5)
    queue_joiner_for_test = QueueJoinerForTest(batch_size=batch_size)
    p.steps[-1] = queue_joiner_for_test
    p.steps_as_tuple[-1] = (p.steps_as_tuple[-1][0], queue_joiner_for_test)
    p._refresh_steps()

    p = p.fit(list(range(100)), list(range(100)))

    assert not p[-1].called_queue_joiner


def test_sequential_queued_pipeline_should_fit_transform_without_multiprocessing():
    batch_size = 10
    p = SequentialQueuedPipeline([
        (1, FitTransformCallbackStep(transform_function=lambda di: np.array(di) * 2)),
        (1, FitTransformCallbackStep(transform_function=lambda di: np.array(di) * 2)),
        (1, FitTransformCallbackStep(transform_function=lambda di: np.array(di) * 2)),
        (1, FitTransformCallbackStep(transform_function=lambda di: np.array(di) * 2))
    ], batch_size=batch_size, max_queue_size=5, use_processes=False, use_savers=False)
    queue_joiner_for_test = QueueJoinerForTest(batch_size=batch_size)
    p.steps[-1] = queue_joiner_for_test
    p.steps_as_tuple[-1] = (p.steps_as_tuple[-1][0], queue_joiner_for_test)
    p._refresh_steps()

    p, outputs = p.fit_transform(list(range(100)), list(range(100)))

    assert not p[-1].called_queue_joiner
    assert np.array_equal(outputs, EXPECTED_OUTPUTS)



import os

import numpy as np

from neuraxle.base import ExecutionContext as CX
from neuraxle.hyperparams.distributions import Boolean
from neuraxle.hyperparams.space import HyperparameterSpace, HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import StepClonerForEachDataInput
from neuraxle.steps.misc import FitCallbackStep, TapeCallbackFunction
from neuraxle.steps.numpy import MultiplyByN

HYPE_SPACE = HyperparameterSpace({
    "a__test": Boolean()
})

HYPE_SAMPLE = HyperparameterSamples({
    "a__test": True
})


def test_step_cloner_should_transform():
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    )
    data_inputs = _create_data((2, 2))

    processed_outputs = p.transform(data_inputs)

    assert isinstance(p.steps_as_tuple[0][1], Pipeline)
    assert isinstance(p.steps_as_tuple[1][1], Pipeline)
    assert np.array_equal(processed_outputs, data_inputs * 2)


def test_step_cloner_should_fit_transform():
    # Given
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    )
    data_inputs = _create_data((2, 2))
    expected_outputs = _create_data((2, 2))

    # When
    p, processed_outputs = p.fit_transform(data_inputs, expected_outputs)

    # Then
    assert isinstance(p.steps_as_tuple[0][1], Pipeline)
    assert np.array_equal(p.steps_as_tuple[0][1][0].callback_function.data[0][0], data_inputs[0])
    assert np.array_equal(p.steps_as_tuple[0][1][0].callback_function.data[0][1], expected_outputs[0])

    assert isinstance(p.steps_as_tuple[1][1], Pipeline)
    assert np.array_equal(p.steps_as_tuple[1][1][0].callback_function.data[0][0], data_inputs[1])
    assert np.array_equal(p.steps_as_tuple[1][1][0].callback_function.data[0][1], expected_outputs[1])

    assert np.array_equal(processed_outputs, data_inputs * 2)


def test_step_cloner_should_inverse_transform():
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    )
    data_inputs = _create_data((2, 2))
    expected_outputs = _create_data((2, 2))

    p, processed_outputs = p.fit_transform(data_inputs, expected_outputs)

    assert np.array_equal(processed_outputs, data_inputs * 2)
    inverse_processed_outputs = p.inverse_transform(processed_outputs)
    assert np.array_equal(np.array(inverse_processed_outputs), np.array(data_inputs))


def test_step_cloner_should_set_train():
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    )
    data_inputs = _create_data((2, 2))
    expected_outputs = _create_data((2, 2))
    p, processed_outputs = p.fit_transform(data_inputs, expected_outputs)

    p.set_train(False)

    assert not p.is_train
    assert not p.steps_as_tuple[0][1].is_train
    assert not p.steps_as_tuple[1][1].is_train


def test_step_cloner_should_save_sub_steps(tmpdir):
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    ).with_context(CX(tmpdir))
    data_inputs = _create_data((2, 2))
    expected_outputs = _create_data((2, 2))
    p, processed_outputs = p.fit_transform(data_inputs, expected_outputs)

    p.save(CX(tmpdir), full_dump=True)

    saved_paths = [
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[0]/FitCallbackStep/FitCallbackStep.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[0]/MultiplyByN/MultiplyByN.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[0]/MultiplyByN/MultiplyByN.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[0]/Pipeline[0].joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[1]/FitCallbackStep/FitCallbackStep.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[1]/MultiplyByN/MultiplyByN.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline[1]/Pipeline[1].joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline/FitCallbackStep/FitCallbackStep.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline/MultiplyByN/MultiplyByN.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/Pipeline/Pipeline.joblib'),
        os.path.join(tmpdir, 'StepClonerForEachDataInput/StepClonerForEachDataInput.joblib')
    ]

    for p in saved_paths:
        assert os.path.exists(p)


def test_step_cloner_should_load_sub_steps(tmpdir):
    tape = TapeCallbackFunction()
    p = StepClonerForEachDataInput(
        Pipeline([
            FitCallbackStep(tape),
            MultiplyByN(2)
        ])
    ).with_context(CX(tmpdir))
    data_inputs = _create_data((2, 2))
    expected_outputs = _create_data((2, 2))
    p, _ = p.fit_transform(data_inputs, expected_outputs)

    p.save(CX(tmpdir), full_dump=True)

    loaded_step_cloner = CX(tmpdir).load('StepClonerForEachDataInput')
    assert isinstance(loaded_step_cloner.wrapped, Pipeline)
    assert len(loaded_step_cloner.steps_as_tuple) == len(data_inputs)
    assert isinstance(loaded_step_cloner.steps_as_tuple[0][1], Pipeline)
    assert isinstance(loaded_step_cloner.steps_as_tuple[1][1], Pipeline)


def _create_data(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    return data_inputs



import copy
from typing import Callable, List

import pytest
from neuraxle.base import CX, synchroneous_flow_method
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.data.vanilla import (DEFAULT_CLIENT, DEFAULT_PROJECT,
                                           AutoMLContext, BaseDataclass,
                                           ClientDataclass,
                                           HyperparamsRepository,
                                           ProjectDataclass, RootDataclass,
                                           RoundDataclass, ScopedLocation,
                                           VanillaHyperparamsRepository,
                                           as_named_odict, from_json, to_json)
from sklearn.metrics import median_absolute_error
from testing.metaopt.test_automl_dataclasses import (ALL_DATACLASSES,
                                                     SOME_CLIENT_DATACLASS,
                                                     SOME_FULL_SCOPED_LOCATION,
                                                     SOME_PROJECT_DATACLASS,
                                                     SOME_ROOT_DATACLASS,
                                                     SOME_ROUND_DATACLASS)

TmpDir = str


def vanilla_repo_ctor(tmpdir: TmpDir = None) -> AutoMLContext:
    return AutoMLContext.from_context(CX())


def disk_repo_ctor(tmpdir: TmpDir = None) -> AutoMLContext:
    cx = CX()
    tmpdir = tmpdir or cx.get_new_cache_folder()
    return AutoMLContext.from_context(cx, repo=HyperparamsOnDiskRepository(tmpdir))


CX_WITH_REPO_CTORS: List[Callable[[TmpDir], AutoMLContext]] = [vanilla_repo_ctor, disk_repo_ctor]


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_auto_ml_context_loc_stays_the_same(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext]
):
    context = cx_repo_ctor(tmpdir)

    c0 = context.push_attr(RootDataclass())
    c1 = c0.push_attr(ProjectDataclass(project_name=DEFAULT_PROJECT))
    c2 = c1.push_attr(ClientDataclass(client_name=DEFAULT_CLIENT))

    assert c0.loc.as_list() == []
    assert c1.loc.as_list() == [DEFAULT_PROJECT]
    assert c2.loc.as_list() == [DEFAULT_PROJECT, DEFAULT_CLIENT]


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_context_changes_independently_once_copied(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext]
):
    cx = cx_repo_ctor(tmpdir)

    copied_cx: AutoMLContext = cx.copy().push_attr(
        ProjectDataclass(project_name=DEFAULT_PROJECT))

    assert copied_cx.loc.as_list() == [DEFAULT_PROJECT]
    assert cx.loc.as_list() == []


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_logger_level_works_with_multiple_levels(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext]
):
    c0 = cx_repo_ctor(tmpdir)

    c0.flow.log("c0.flow.log: begin")
    c1 = c0.push_attr(ProjectDataclass(project_name=DEFAULT_PROJECT))
    c1.flow.log("c1.flow.log: begin")
    c1.flow.log("c1.flow.log: some work being done from within c1")
    c1.flow.log("c1.flow.log: end")
    c0.flow.log("c0.flow.log: end")

    l0: str = c0.read_scoped_log()
    l1: str = c1.read_scoped_log()

    assert l0 != l1
    assert len(l0) > len(l1)
    assert "c0" in l0
    assert "c0" not in l1
    assert "c1" in l0
    assert "c1" in l1
    assert c1.loc != c0.loc


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_hyperparams_repository_has_default_client_project(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext]
):
    cx = cx_repo_ctor(tmpdir)
    loc: ScopedLocation = ScopedLocation.default().with_dc(SOME_ROUND_DATACLASS)

    restored_client = cx.repo.load(loc.popped(), deep=True)
    restored_project = cx.repo.load(loc.popped().popped(), deep=True)

    cx.repo.save(copy.deepcopy(SOME_ROUND_DATACLASS), scope=loc, deep=True)
    restored_round = cx.repo.load(loc, deep=True)

    assert restored_round == SOME_ROUND_DATACLASS
    assert restored_project.shallow() == SOME_PROJECT_DATACLASS.shallow()
    assert restored_client.get_id() == SOME_CLIENT_DATACLASS.get_id()


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
@pytest.mark.parametrize('_dataclass', ALL_DATACLASSES)
def test_hyperparams_repository_loads_stored_scoped_info(
    tmpdir, cx_repo_ctor: Callable[[str], AutoMLContext], _dataclass: BaseDataclass
):
    loc: ScopedLocation = SOME_FULL_SCOPED_LOCATION.at_dc(_dataclass)
    cx: AutoMLContext = cx_repo_ctor(tmpdir).with_loc(loc)
    repo: HyperparamsRepository = cx.repo
    repo.save(copy.deepcopy(SOME_ROOT_DATACLASS), scope=loc, deep=True)

    restored_dataclass_deep = repo.load(loc, deep=True)
    restored_dataclass_shallow = repo.load(loc, deep=False)

    assert restored_dataclass_deep == _dataclass
    assert restored_dataclass_shallow == _dataclass.shallow()


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
@pytest.mark.parametrize('_dataclass', ALL_DATACLASSES[1:])
def test_hyperparams_repository_saves_subsequent_data(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext], _dataclass: BaseDataclass
):
    loc: ScopedLocation = SOME_FULL_SCOPED_LOCATION.at_dc(_dataclass)
    cx: AutoMLContext = cx_repo_ctor(tmpdir).with_loc(loc)
    repo: HyperparamsRepository = cx.repo
    repo.save(copy.deepcopy(SOME_ROOT_DATACLASS), scope=loc, deep=True)
    old_id = _dataclass.get_id()
    next_id = old_id + 1 if isinstance(old_id, int) else old_id + "_next"
    next_dataclass = copy.deepcopy(_dataclass).set_id(next_id)
    next_loc = loc.popped().with_dc(next_dataclass)

    repo.save(next_dataclass, scope=next_loc, deep=False)

    restored_next_dataclass_empty = repo.load(next_loc, deep=True)
    assert restored_next_dataclass_empty == next_dataclass.empty()

    restored_next_dataclass_empty = repo.load(next_loc, deep=False)
    assert restored_next_dataclass_empty == next_dataclass.empty()

    repo.save(next_dataclass, scope=next_loc, deep=True)

    restored_dataclass_shallow = repo.load(next_loc, deep=False)
    assert restored_dataclass_shallow == next_dataclass.shallow()

    restored_dataclass_deep = repo.load(next_loc, deep=True)
    assert restored_dataclass_deep == next_dataclass


@pytest.mark.parametrize('cx_repo_ctor', CX_WITH_REPO_CTORS)
def test_shallow_save_isnt_deep_upon_load(
    tmpdir, cx_repo_ctor: Callable[[TmpDir], AutoMLContext]
):
    zero_loc: ScopedLocation = SOME_FULL_SCOPED_LOCATION.at_dc(SOME_ROUND_DATACLASS)
    cx: AutoMLContext = cx_repo_ctor(tmpdir).with_loc(zero_loc)
    repo: HyperparamsRepository = cx.repo
    next_round_dataclass: RoundDataclass = copy.deepcopy(SOME_ROUND_DATACLASS).set_id(1)
    next_loc = zero_loc.popped().with_dc(next_round_dataclass)

    repo.save(SOME_ROUND_DATACLASS, scope=zero_loc, deep=False)
    repo.save(next_round_dataclass, scope=next_loc, deep=False)
    reloaded_next_round = repo.load(next_loc, deep=True)

    assert reloaded_next_round == next_round_dataclass.empty()



import numpy as np

from neuraxle.pipeline import MiniBatchSequentialPipeline, Joiner
from neuraxle.steps.misc import TransformCallbackStep, TapeCallbackFunction, FitTransformCallbackStep


class MultiplyBy2TransformCallbackStep(TransformCallbackStep):
    def transform(self, data_inputs):
        TransformCallbackStep.transform(self, data_inputs)

        return list(np.array(data_inputs) * 2)


def test_mini_batch_sequential_pipeline_should_transform_steps_sequentially_for_each_barrier_for_each_batch():
    # Given
    tape1 = TapeCallbackFunction()
    tape2 = TapeCallbackFunction()
    tape3 = TapeCallbackFunction()
    tape4 = TapeCallbackFunction()
    p = MiniBatchSequentialPipeline([
        MultiplyBy2TransformCallbackStep(tape1, ["1"]),
        MultiplyBy2TransformCallbackStep(tape2, ["2"]),
        Joiner(batch_size=10),
        MultiplyBy2TransformCallbackStep(tape3, ["3"]),
        MultiplyBy2TransformCallbackStep(tape4, ["4"]),
        Joiner(batch_size=10)
    ])

    # When
    outputs = p.transform(list(range(20)))

    # Then
    assert outputs == [0, 16, 32, 48, 64, 80, 96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 256, 272, 288, 304]

    assert tape1.data == [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]]
    assert tape1.name_tape == ["1", "1"]

    assert tape2.data == [[0, 2, 4, 6, 8, 10, 12, 14, 16, 18], [20, 22, 24, 26, 28, 30, 32, 34, 36, 38]]
    assert tape2.name_tape == ["2", "2"]

    assert tape3.data == [[0, 4, 8, 12, 16, 20, 24, 28, 32, 36], [40, 44, 48, 52, 56, 60, 64, 68, 72, 76]]
    assert tape3.name_tape == ["3", "3"]

    assert tape4.data == [[0, 8, 16, 24, 32, 40, 48, 56, 64, 72], [80, 88, 96, 104, 112, 120, 128, 136, 144, 152]]
    assert tape4.name_tape == ["4", "4"]


class MultiplyBy2FitTransformCallbackStep(FitTransformCallbackStep):
    def fit_transform(self, data_inputs, expected_outputs=None):
        FitTransformCallbackStep.fit_transform(self, data_inputs, expected_outputs)

        return self, list(np.array(data_inputs) * 2)


def test_mini_batch_sequential_pipeline_should_fit_transform_steps_sequentially_for_each_barrier_for_each_batch():
    # Given
    tape1 = TapeCallbackFunction()
    tape1_fit = TapeCallbackFunction()
    tape2 = TapeCallbackFunction()
    tape2_fit = TapeCallbackFunction()
    tape3 = TapeCallbackFunction()
    tape3_fit = TapeCallbackFunction()
    tape4 = TapeCallbackFunction()
    tape4_fit = TapeCallbackFunction()
    p = MiniBatchSequentialPipeline([
        MultiplyBy2FitTransformCallbackStep(tape1, tape1_fit, ["1"]),
        MultiplyBy2FitTransformCallbackStep(tape2, tape2_fit, ["2"]),
        Joiner(batch_size=10),
        MultiplyBy2FitTransformCallbackStep(tape3, tape3_fit, ["3"]),
        MultiplyBy2FitTransformCallbackStep(tape4, tape4_fit, ["4"]),
        Joiner(batch_size=10)
    ])

    # When
    p, outputs = p.fit_transform(list(range(20)), list(range(20)))

    # Then
    assert outputs == [0, 16, 32, 48, 64, 80, 96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 256, 272, 288, 304]

    assert tape1.data == [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]]
    assert tape1_fit.data == [([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([10, 11, 12, 13, 14, 15, 16, 17, 18, 19], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])]
    assert tape1.name_tape == ["1", "1"]

    assert tape2.data == [[0, 2, 4, 6, 8, 10, 12, 14, 16, 18], [20, 22, 24, 26, 28, 30, 32, 34, 36, 38]]
    assert tape2_fit.data == [([0, 2, 4, 6, 8, 10, 12, 14, 16, 18], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([20, 22, 24, 26, 28, 30, 32, 34, 36, 38], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])]
    assert tape2.name_tape == ["2", "2"]

    assert tape3.data == [[0, 4, 8, 12, 16, 20, 24, 28, 32, 36], [40, 44, 48, 52, 56, 60, 64, 68, 72, 76]]
    assert tape3_fit.data == [([0, 4, 8, 12, 16, 20, 24, 28, 32, 36], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([40, 44, 48, 52, 56, 60, 64, 68, 72, 76], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])]
    assert tape3.name_tape == ["3", "3"]

    assert tape4.data == [[0, 8, 16, 24, 32, 40, 48, 56, 64, 72], [80, 88, 96, 104, 112, 120, 128, 136, 144, 152]]
    assert tape4_fit.data == [([0, 8, 16, 24, 32, 40, 48, 56, 64, 72], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([80, 88, 96, 104, 112, 120, 128, 136, 144, 152], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19])]
    assert tape4.name_tape == ["4", "4"]

def test_minibatch_sequential_pipeline_change_batch_size_works():
    tape1 = TapeCallbackFunction()
    tape1_fit = TapeCallbackFunction()
    tape2 = TapeCallbackFunction()
    tape2_fit = TapeCallbackFunction()

    p = MiniBatchSequentialPipeline([
        MultiplyBy2FitTransformCallbackStep(tape1, tape1_fit, ["1"]),
        Joiner(batch_size=10),
        MultiplyBy2FitTransformCallbackStep(tape2, tape2_fit, ["2"]),
        Joiner(batch_size=10)
    ])

    # When
    p, outputs = p.fit_transform(list(range(20)), list(range(20)))
    p.set_batch_size(5)
    p, outputs = p.fit_transform(list(range(20,30)), list(range(20,30)))

    # Then

    assert tape1.data == [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                          [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]
    assert tape1_fit.data == [([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([10, 11, 12, 13, 14, 15, 16, 17, 18, 19], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                              ([20, 21, 22, 23, 24], [20, 21, 22, 23, 24]),
                              ([25, 26, 27, 28, 29], [25, 26, 27, 28, 29])]
    assert tape1.name_tape == ["1", "1", "1", "1"]

    assert tape2.data == [[0, 2, 4, 6, 8, 10, 12, 14, 16, 18], [20, 22, 24, 26, 28, 30, 32, 34, 36, 38],
                          [40, 42, 44, 46, 48], [50, 52, 54, 56, 58]]
    assert tape2_fit.data == [([0, 2, 4, 6, 8, 10, 12, 14, 16, 18], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                              ([20, 22, 24, 26, 28, 30, 32, 34, 36, 38], [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]),
                              ([40, 42, 44, 46, 48], [20, 21, 22, 23, 24]),
                              ([50, 52, 54, 56, 58], [25, 26, 27, 28, 29])]
    assert tape2.name_tape == ["2", "2", "2", "2"]





"""
Tests for Pipelines
========================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

from typing import Generic
import numpy as np
import pytest

from neuraxle.base import (_FittableStep, _HasChildrenMixin, BaseStep, BaseTransformer,
                           CX, NonTransformableMixin)
from neuraxle.hyperparams.distributions import RandInt, LogUniform
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.pipeline import Pipeline
from neuraxle.steps.misc import TransformCallbackStep, TapeCallbackFunction
from neuraxle.steps.numpy import NumpyTranspose
from neuraxle.union import Identity, AddFeatures, ModelStacking
from testing.mocks.step_mocks import SomeStep, AN_INPUT, AN_EXPECTED_OUTPUT

steps_lists = [
    [("just_one_step", SomeStep())],
    [
        ("some_step_1", SomeStep()),
        ("some_step_2", SomeStep()),
        ("some_step_3", SomeStep())
    ]
]


@pytest.mark.parametrize("steps_list", steps_lists)
def test_pipeline_fit_transform(steps_list):
    data_input_ = [AN_INPUT]
    expected_output_ = [AN_EXPECTED_OUTPUT]
    p = Pipeline(steps_list)

    p, result = p.fit_transform(data_input_, expected_output_)

    assert tuple(result) == tuple(expected_output_)


@pytest.mark.parametrize("steps_list", steps_lists)
def test_pipeline_fit_then_transform(steps_list):
    data_input_ = [AN_INPUT]
    expected_output_ = [AN_EXPECTED_OUTPUT]
    p = Pipeline(steps_list)

    p = p.fit(data_input_, expected_output_)
    result = p.transform(data_input_)

    assert tuple(result) == tuple(expected_output_)


def test_pipeline_slicing_before():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])

    r = p["b":]

    assert "a" not in r
    assert "b" in r
    assert "c" in r


def test_pipeline_slicing_after():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])

    r = p[:"c"]

    assert "a" in r
    assert "b" in r
    assert "c" not in r


def test_pipeline_slicing_both():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])

    r = p["b":"c"]

    assert "a" not in r
    assert "b" in r
    assert "c" not in r


def test_pipeline_set_one_hyperparam_level_one_flat():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])

    p.set_hyperparams({
        "a__learning_rate": 7
    })

    assert p["a"].hyperparams.to_flat_dict()["learning_rate"] == 7
    assert p["b"].hyperparams.to_flat_dict() == dict()
    assert p["c"].hyperparams.to_flat_dict() == dict()


def test_pipeline_set_one_hyperparam_level_one_dict():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])

    p.set_hyperparams({
        "b": {
            "learning_rate": 7
        }
    })

    assert p["a"].hyperparams == dict()
    assert p["b"].hyperparams["learning_rate"] == 7
    assert p["c"].hyperparams == dict()


def test_pipeline_set_one_hyperparam_level_two_flat():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", Pipeline([
            ("a", SomeStep()),
            ("b", SomeStep()),
            ("c", SomeStep())
        ])),
        ("c", SomeStep())
    ])

    p.set_hyperparams({
        "b__a__learning_rate": 7
    })
    print(p.get_hyperparams())

    assert p["b"]["a"].hyperparams["learning_rate"] == 7
    assert p["b"]["c"].hyperparams.to_flat_dict() == dict()
    assert p["b"].hyperparams.to_flat_dict() == {'a__learning_rate': 7}
    assert p["c"].hyperparams.to_flat_dict() == dict()


def test_pipeline_set_one_hyperparam_level_two_dict():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", Pipeline([
            ("a", SomeStep()),
            ("b", SomeStep()),
            ("c", SomeStep())
        ])),
        ("c", SomeStep())
    ])

    p.set_hyperparams({
        "b": {
            "a": {
                "learning_rate": 7
            },
            "learning_rate": 9
        }
    })
    print(p.get_hyperparams())

    assert p["b"]["a"].get_hyperparams()["learning_rate"] == 7
    assert p["b"]["c"].get_hyperparams() == dict()
    assert p["b"].get_hyperparams()["learning_rate"] == 9
    assert p["c"].get_hyperparams() == dict()


def test_pipeline_update_hyperparam_level_one_flat():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])
    p.set_hyperparams({
        "a__learning_rate": 7,
        "a__other_hp": 8
    })

    p.update_hyperparams({
        "a__learning_rate": 0.01
    })

    assert p["a"].hyperparams["learning_rate"] == 0.01
    assert p["a"].hyperparams["other_hp"] == 8
    assert p["b"].hyperparams == dict()
    assert p["c"].hyperparams == dict()


def test_pipeline_update_hyperparam_level_one_dict():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", SomeStep()),
        ("c", SomeStep())
    ])
    p.set_hyperparams({"b": {"learning_rate": 7, "other_hp": 8}})

    p.update_hyperparams({"b": {"learning_rate": 0.01}})

    assert p["b"].hyperparams["learning_rate"] == 0.01
    assert p["b"].hyperparams["other_hp"] == 8
    assert p["a"].hyperparams == dict()
    assert p["c"].hyperparams == dict()


def test_pipeline_update_hyperparam_level_two_flat():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", Pipeline([
            ("a", SomeStep()),
            ("b", SomeStep()),
            ("c", SomeStep())
        ])),
        ("c", SomeStep())
    ])
    p.set_hyperparams({
        "b__a__learning_rate": 7,
        "b__a__other_hp": 8,
    })

    p.update_hyperparams({
        "b__a__learning_rate": 0.01
    })

    assert p["b"]["a"].hyperparams["learning_rate"] == 0.01
    assert p["b"]["a"].hyperparams["other_hp"] == 8
    assert p["b"]["c"].hyperparams == dict()
    assert p["b"].hyperparams.to_flat_dict() == {
        'a__learning_rate': 0.01,
        'a__other_hp': 8
    }
    assert p["c"].hyperparams == dict()


def test_pipeline_update_hyperparam_level_two_dict():
    p = Pipeline([
        ("a", SomeStep()),
        ("b", Pipeline([
            ("a", SomeStep()),
            ("b", SomeStep()),
            ("c", SomeStep())
        ])),
        ("c", SomeStep())
    ])
    p.set_hyperparams({"b": {"a": {"learning_rate": 7, "other_hp": 8}, "learning_rate": 9}})

    p.update_hyperparams({"b": {"a": {"learning_rate": 0.01}}})

    assert p["b"]["a"].hyperparams["learning_rate"] == 0.01
    assert p["b"]["a"].hyperparams["other_hp"] == 8
    assert p["b"]["c"].hyperparams == dict()
    assert p["b"].hyperparams["learning_rate"] == 9
    assert p["c"].hyperparams == dict()


def test_pipeline_simple_inverse_transform():
    expected_tape = ["1", "2", "3", "4", "4", "3", "2", "1"]
    tape = TapeCallbackFunction()

    p = Pipeline([
        Identity(),
        TransformCallbackStep(tape.callback, ["1"]),
        TransformCallbackStep(tape.callback, ["2"]),
        TransformCallbackStep(tape.callback, ["3"]),
        TransformCallbackStep(tape.callback, ["4"]),
        Identity()
    ])

    p, _ = p.fit_transform(np.ones((1, 1)))
    p.inverse_transform(np.ones((1, 1)))

    assert expected_tape == tape.get_name_tape()


def test_pipeline_nested_inverse_transform():
    expected_tape = ["1", "2", "3", "4", "5", "6", "7", "7", "6", "5", "4", "3", "2", "1"]
    tape = TapeCallbackFunction()

    p = Pipeline([
        Identity(),
        TransformCallbackStep(tape.callback, ["1"]),
        TransformCallbackStep(tape.callback, ["2"]),
        Pipeline([
            Identity(),
            TransformCallbackStep(tape.callback, ["3"]),
            TransformCallbackStep(tape.callback, ["4"]),
            TransformCallbackStep(tape.callback, ["5"]),
            Identity()
        ]),
        TransformCallbackStep(tape.callback, ["6"]),
        TransformCallbackStep(tape.callback, ["7"]),
        Identity()
    ])

    p, _ = p.fit_transform(np.ones((1, 1)))  # will add range(1, 8) to tape.

    p.inverse_transform(np.ones((1, 1)))  # will add reversed(range(1, 8)) to tape.

    print(expected_tape)
    print(tape.get_name_tape())
    assert expected_tape == tape.get_name_tape()


def test_pipeline_nested_inverse_transform_without_identities():
    """
    This test was required for a strange bug at the border of the pipelines
    that happened when the identities were not used.
    """
    expected_tape = ["1", "2", "3", "4", "5", "6", "7", "7", "6", "5", "4", "3", "2", "1"]
    tape = TapeCallbackFunction()

    p = Pipeline([
        TransformCallbackStep(tape.callback, ["1"]),
        TransformCallbackStep(tape.callback, ["2"]),
        Pipeline([
            TransformCallbackStep(tape.callback, ["3"]),
            TransformCallbackStep(tape.callback, ["4"]),
            TransformCallbackStep(tape.callback, ["5"]),
        ]),
        TransformCallbackStep(tape.callback, ["6"]),
        TransformCallbackStep(tape.callback, ["7"]),
    ])

    p, _ = p.fit_transform(np.ones((1, 1)))  # will add range(1, 8) to tape.

    p.inverse_transform(np.ones((1, 1)))  # will add reversed(range(1, 8)) to tape, calling inverse_transforms.

    print(expected_tape)
    print(tape.get_name_tape())
    assert expected_tape == tape.get_name_tape()


def test_hyperparam_space():
    p = Pipeline([
        AddFeatures([
            SomeStep(hyperparams_space=HyperparameterSpace({"n_components": RandInt(1, 5)})),
            SomeStep(hyperparams_space=HyperparameterSpace({"n_components": RandInt(1, 5)}))
        ]),
        ModelStacking([
            SomeStep(hyperparams_space=HyperparameterSpace({"n_estimators": RandInt(1, 1000)})),
            SomeStep(hyperparams_space=HyperparameterSpace({"n_estimators": RandInt(1, 1000)})),
            SomeStep(hyperparams_space=HyperparameterSpace({"max_depth": RandInt(1, 100)})),
            SomeStep(hyperparams_space=HyperparameterSpace({"max_depth": RandInt(1, 100)}))
        ],
            joiner=NumpyTranspose(),
            judge=SomeStep(hyperparams_space=HyperparameterSpace({"alpha": LogUniform(0.1, 10.0)}))
        )
    ])

    rvsed = p.get_hyperparams_space()
    p.set_hyperparams(rvsed)

    hyperparams = p.get_hyperparams()
    flat_hyperparams_keys = hyperparams.to_flat_dict().keys()

    assert 'AddFeatures' in hyperparams
    assert 'SomeStep' in hyperparams["AddFeatures"]
    assert "n_components" in hyperparams["AddFeatures"]["SomeStep"]
    assert 'SomeStep1' in hyperparams["AddFeatures"]
    assert "n_components" in hyperparams["AddFeatures"]["SomeStep1"]

    assert 'ModelStacking' in hyperparams
    assert 'SomeStep' in hyperparams["ModelStacking"]
    assert 'n_estimators' in hyperparams["ModelStacking"]["SomeStep"]
    assert 'SomeStep1' in hyperparams["ModelStacking"]
    assert 'n_estimators' in hyperparams["ModelStacking"]["SomeStep1"]
    assert 'SomeStep2' in hyperparams["ModelStacking"]
    assert 'max_depth' in hyperparams["ModelStacking"]["SomeStep2"]
    assert 'SomeStep3' in hyperparams["ModelStacking"]
    assert 'max_depth' in hyperparams["ModelStacking"]["SomeStep3"]

    assert 'AddFeatures__SomeStep1__n_components' in flat_hyperparams_keys
    assert 'AddFeatures__SomeStep__n_components' in flat_hyperparams_keys
    assert 'ModelStacking__SomeStep__n_estimators' in flat_hyperparams_keys
    assert 'ModelStacking__SomeStep1__n_estimators' in flat_hyperparams_keys
    assert 'ModelStacking__SomeStep2__max_depth' in flat_hyperparams_keys
    assert 'ModelStacking__SomeStep3__max_depth' in flat_hyperparams_keys


def test_pipeline_setup_incrementally():
    class SomeStepThatFits(NonTransformableMixin, BaseStep):
        def __init__(self):
            BaseStep.__init__(self)
            self.has_fitted = False

        def fit(self, data_inputs, expected_outputs=None) -> _FittableStep:
            self.has_fitted = True
            return self

    class StepWithSensitiveSetup(Identity):
        """ Asserts that step given in argument has fitted before performing setup"""

        def __init__(self):
            Identity.__init__(self)

        def setup(self, context: CX = None) -> BaseTransformer:
            assert some_step.has_fitted is True
            assert some_step2.has_fitted is False
            return self

    some_step = SomeStepThatFits()
    some_step2 = SomeStepThatFits()

    p = Pipeline([some_step,
                  StepWithSensitiveSetup(),
                  some_step2])

    p.fit_transform(None, None)


def test_subtyping_of_pipeline_works_correctly():
    p: Pipeline[Identity] = Pipeline([Identity(), Identity()])

    assert issubclass(Pipeline, Generic)
    assert isinstance(p, _HasChildrenMixin)
    assert isinstance(p[-1], Identity)



import sys
import typing
from typing import Generic, TypeVar

import pytest
from neuraxle.base import BaseService, MetaStep, NonFittableMixin
from neuraxle.pipeline import Pipeline
from neuraxle.union import Identity

from testing.mocks.step_mocks import SomeMetaStepWithHyperparams


def test_metastepmixin_set_train_should_set_train_to_false():
    p = MetaStep(Pipeline([
        Identity()
    ]))

    p.set_train(False)

    assert not p.is_train
    assert not p.wrapped[0].is_train
    assert not p.wrapped.is_train


def test_metastepmixin_set_train_should_set_train_to_true():
    p = MetaStep(Pipeline([
        Identity()
    ]))

    assert p.is_train
    assert p.wrapped[0].is_train
    assert p.wrapped.is_train


def test_basestep_str_representation_works_correctly():
    output = str(SomeMetaStepWithHyperparams())
    assert output == "SomeMetaStepWithHyperparams(SomeStepWithHyperparams(name='MockStep'), name='SomeMetaStepWithHyperparams')"


def test_subtyping_of_metastep_works_correctly():
    some_step: MetaStep[Identity] = MetaStep(Identity())

    assert issubclass(MetaStep, Generic)
    assert isinstance(some_step, MetaStep)
    assert isinstance(some_step.get_step(), Identity)


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Python 3.8 or more needed")
def test_typable_mixin_can_hold_type_annotation(tmpdir):
    # Testing the type annotation "MetaStep[MyService]":
    wrapped_service: MetaStep[Identity] = MetaStep(Identity())

    g: Generic = wrapped_service.__orig_bases__[-1]
    assert isinstance(wrapped_service.get_step(), g.__parameters__[0].__bound__)
    bt: TypeVar = typing.get_args(g)[0]
    assert isinstance(wrapped_service.get_step(), bt.__bound__)

    assert isinstance(wrapped_service.get_step(), Identity)
    assert isinstance(wrapped_service.get_step(), NonFittableMixin)
    assert isinstance(wrapped_service.get_step(), BaseService)



import numpy as np
import pytest

from neuraxle.steps.numpy import OneHotEncoder


@pytest.mark.parametrize("n_dims", [1, 2, 3])
@pytest.mark.parametrize("no_columns", [10])
def test_one_hot_encode_should_encode_data_inputs(n_dims, no_columns):
    one_hot_encode = OneHotEncoder(nb_columns=no_columns, name='one_hot')
    data_shape = list(range(100, 200))[:n_dims]
    data_inputs = np.random.randint(low=no_columns, size=data_shape)
    data_inputs[0] = 0
    data_inputs[1] = no_columns - 1
    data_inputs[-2] = -1  # or nan or inf.

    outputs = one_hot_encode.transform(data_inputs)

    assert outputs.shape[-1] == no_columns
    assert ((outputs == 1) | (outputs == 0)).all()

    if n_dims >= 2:
        assert (outputs[0, ..., 0] == 1).all()
        assert (outputs[1, ..., -1] == 1).all()
        assert (outputs[-2, ...] == 0).all()



import numpy as np

from neuraxle.data_container import DACT, ListDataContainer


def test_data_container_iter_method_should_iterate_with_none_ids():
    data_container = DACT(
        di=np.array(list(range(100))),
        eo=np.array(list(range(100, 200)))
    ).set_ids(None)

    for i, (_id, data_input, expected_outputs) in enumerate(data_container):
        assert _id == i
        assert data_input == i
        assert expected_outputs == i + 100


def test_data_container_iter_method_should_iterate_with_none_expected_outputs():
    data_container = DACT(
        ids=[str(i) for i in range(100)],
        data_inputs=np.array(list(range(100))),
        expected_outputs=None
    )

    for i, (_, data_input, expected_outputs) in enumerate(data_container):
        assert data_input == i
        assert expected_outputs is None


def test_data_container_len_method_should_return_data_inputs_len():
    data_container = DACT.from_di(np.array(list(range(100))))

    assert len(data_container) == 100


def test_data_container_should_iterate_through_data_using_minibatches():
    data_container = DACT(
        ids=[str(i) for i in range(100)],
        data_inputs=np.array(list(range(100))),
        expected_outputs=np.array(list(range(100, 200)))
    )

    batches = []
    for b in data_container.minibatches(batch_size=10):
        batches.append(b)

    for i, batch in enumerate(batches):
        assert np.array_equal(np.array(batch.data_inputs), np.array(list(range(i * 10, (i * 10) + 10))))
        assert np.array_equal(
            np.array(batch.expected_outputs),
            np.array(list(range((i * 10) + 100, (i * 10) + 100 + 10)))
        )


def test_list_data_container_concat():
    # Given
    data_container = ListDataContainer(
        ids=[str(i) for i in range(100)],
        data_inputs=np.array(list(range(100))),
        expected_outputs=np.array(list(range(100, 200)))
    )

    # When
    data_container.concat(DACT(
        ids=[str(i) for i in range(100, 200)],
        data_inputs=np.array(list(range(100, 200))),
        expected_outputs=np.array(list(range(200, 300)))
    ))

    # Then
    assert np.array_equal(np.array(data_container.ids), np.array(list(range(0, 200))).astype(np.str))

    expected_data_inputs = np.array(list(range(0, 200))).astype(np.int)
    actual_data_inputs = np.array(data_container.data_inputs).astype(np.int)
    assert np.array_equal(actual_data_inputs, expected_data_inputs)

    expected_expected_outputs = np.array(list(range(100, 300))).astype(np.int)
    assert np.array_equal(np.array(data_container.expected_outputs).astype(np.int), expected_expected_outputs)



"""
Boston Housing Regression with Meta Optimization
================================================

This is an automatic machine learning example. It is more sophisticated than the other simple regression example.
Not only a pipeline is defined, but also an hyperparameter space is defined for the pipeline. Then, a random search is
performed to find the best possible combination of hyperparameters by sampling randomly in the hyperparameter space.

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np
from sklearn.cluster import KMeans
from sklearn.datasets import load_boston
from sklearn.decomposition import PCA, FastICA
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.linear_model import Ridge
from sklearn.metrics import r2_score, mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.utils import shuffle

from neuraxle.hyperparams.distributions import RandInt, LogUniform, Boolean
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.auto_ml import AutoML, ValidationSplitter
from neuraxle.metaopt.callbacks import MetricCallback
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import NumpyTranspose
from neuraxle.steps.sklearn import SKLearnWrapper
from neuraxle.union import AddFeatures, ModelStacking


def main(tmpdir):
    boston = load_boston()
    X, y = shuffle(boston.data, boston.target, random_state=13)
    X = X.astype(np.float32)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, shuffle=False)

    # Note that the hyperparameter spaces are defined here during the pipeline definition, but it could be already set
    # within the classes ar their definition if using custom classes, or also it could be defined after declaring the
    # pipeline using a flat dict or a nested dict.

    p = Pipeline([
        AddFeatures([
            SKLearnWrapper(
                PCA(n_components=2),
                HyperparameterSpace({"n_components": RandInt(1, 3)})
            ),
            SKLearnWrapper(
                FastICA(n_components=2),
                HyperparameterSpace({"n_components": RandInt(1, 3)})
            ),
        ]),
        ModelStacking([
            SKLearnWrapper(
                GradientBoostingRegressor(),
                HyperparameterSpace({
                    "n_estimators": RandInt(50, 300), "max_depth": RandInt(1, 4),
                    "learning_rate": LogUniform(0.07, 0.7)
                })
            ),
            SKLearnWrapper(
                KMeans(),
                HyperparameterSpace({"n_clusters": RandInt(5, 10)})
            ),
        ],
            joiner=NumpyTranspose(),
            judge=SKLearnWrapper(
                Ridge(),
                HyperparameterSpace({"alpha": LogUniform(0.7, 1.4), "fit_intercept": Boolean()})),
        )
    ])

    print("Meta-fitting on train:")
    auto_ml = AutoML(
        p,
        validation_splitter=ValidationSplitter(0.20),
        n_trials=10,
        epochs=1,  # 1 epoch here due to using sklearn models that just fit once.
        callbacks=[MetricCallback('mse', metric_function=mean_squared_error, higher_score_is_better=False)],
    )

    fitted_random_search = auto_ml.fit(X_train, y_train)
    print("")

    print("Transforming train and test:")
    y_train_predicted = fitted_random_search.predict(X_train)
    y_test_predicted = fitted_random_search.predict(X_test)

    print("")

    print("Evaluating transformed train:")
    score_transform = r2_score(y_train_predicted, y_train)
    print('R2 regression score:', score_transform)

    print("")

    print("Evaluating transformed test:")
    score_test = r2_score(y_test_predicted, y_test)
    print('R2 regression score:', score_test)


if __name__ == "__main__":
    main('cache')
